
/*
Copyright 2007-2011 Kim A. Brandt <kimabrandt@gmx.de>

This file is part of yarip.

Yarip is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

Yarip is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with yarip; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

var yarip = {
    get VERSION() { return "0.2.5.3"; },
    get FILE() { return "yarip.xml"; },
    get CHARSET() { return "UTF-8"; },

    get MODE_PAGE() { return 0; },
    get MODE_FQDN() { return 1; },
    get MODE_SLD() { return 2; },

    get TYPE_WHITELIST() { return 0; },
    get TYPE_BLACKLIST() { return 1; },
    get TYPE_ATTRIBUTE() { return 2; },

    get DATA_TYPE_BOOLEAN() { return 0; },
    get DATA_TYPE_INTEGER() { return 1; },
    get DATA_TYPE_STRING() { return 2; },
    get DATA_TYPE_RESET() { return 3; },

    get INCREMENT_NOT_FOUND() { return 0; },
    get INCREMENT_IGNORED() { return 1; },
    get INCREMENT_FOUND() { return 2; },

    get STATUS_UNKNOWN() { return 0; },
    get STATUS_WHITELISTED() { return 1; },
    get STATUS_BLACKLISTED() { return 2; },
    get STATUS_REDIRECTED() { return 3; },

    get PREF_VERSION() { return "extensions.yarip.version"; },
    get PREF_ENABLED() { return "extensions.yarip.enabled.value"; },
    get PREF_FLICKER() { return "extensions.yarip.noFlicker.value"; },
    get PREF_MODE() { return "extensions.yarip.mode.value"; },
    get PREF_INDEX() { return "extensions.yarip.alwaysUseIndex.value"; },
    get PREF_ELEMENTS() { return "extensions.yarip.elementsInContext.value"; },
    get PREF_EXCLUSIVE() { return "extensions.yarip.exclusiveOnCreation.value"; },
    get PREF_TEMPLATES() { return "extensions.yarip.templatesList.value"; },
    get PREF_PRIVATE() { return "extensions.yarip.privateBrowsing.value"; },
    get PREF_PURGE() { return "extensions.yarip.purgeInnerHTML.value"; },
    get PREF_SCHEMES() { return "extensions.yarip.schemesRegExp.value"; },

    get urlRegExp() { return URL_RE; },
    get uriSimpleRegExp() { return URI_SIMPLE_RE; },
    get noSchemeRegExp() { return NO_SCHEME_RE; },
    get domainRegExp() { return DOMAIN_RE; },
    get findSldRegExp() { return FIND_SLD_RE; },
    get findTldRegExp() { return FIND_TLD_RE; },
    get outlineRegExp() { return /^outline-color:\ rgb\((0|153),\ 0,\ (0|153)\);\ outline-style:\ solid;\ outline-width:\ 3px;$/; },
    get endOutlineRegExp() { return /outline-color:\ rgb\((0|153),\ 0,\ (0|153)\);\ outline-style:\ solid;\ outline-width:\ 3px;$/; },

    enabled: null,
    noFlicker: null,
    mode: null,

    monitorDialogues: [],
    pageDialog: null,

    alwaysUseIndex: false,
    elementsInContext: 4,
    exclusiveOnCreation: false,
    templates: [],
    privateBrowsing: false,
    purgeInnerHTML: false,
    schemesRegExp: null,

    map: null,
    undoObj: {},
    pageCreated: false,
    knownAddressObj: {}
}

const Cc = Components.classes;
const Ci = Components.interfaces;
const Cr = Components.results;
const Cu = Components.utils;

Cu.import("resource://gre/modules/XPCOMUtils.jsm");

var SSL = Cc["@mozilla.org/moz/jssubscript-loader;1"].getService(Ci.mozIJSSubScriptLoader);
var IOS = Cc["@mozilla.org/network/io-service;1"].getService(Ci.nsIIOService);
var PB = Cc["@mozilla.org/preferences-service;1"].getService(Ci.nsIPrefBranch);
var CS = Cc["@mozilla.org/consoleservice;1"].getService(Ci.nsIConsoleService);
var OS = Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService);
var IDNS = Cc["@mozilla.org/network/idn-service;1"].getService(Ci.nsIIDNService);
var WP = Cc['@mozilla.org/docloaderservice;1'].getService(Ci.nsIWebProgress);

SSL.loadSubScript("chrome://yarip/content/aatree.js");
SSL.loadSubScript("chrome://yarip/content/map.js");
SSL.loadSubScript("chrome://yarip/content/page.js");
SSL.loadSubScript("chrome://yarip/content/list.js");
SSL.loadSubScript("chrome://yarip/content/item.js");
SSL.loadSubScript("chrome://yarip/content/uri.js");
SSL.loadSubScript("chrome://yarip/content/replace.js");

yarip.toggleEnabled = function(enabled)
{
    if (enabled == this.enabled) return false;

    if (enabled != true && enabled != false) enabled = !this.enabled;
    this.enabled = enabled;

    if (enabled) this.toggleNoFlicker(this.getValue(this.PREF_FLICKER, false, this.DATA_TYPE_BOOLEAN), null, true);
    else this.toggleNoFlicker(false, true, true);

    this.setValue(this.PREF_ENABLED, enabled, this.DATA_TYPE_BOOLEAN); // notify
    return true;
}

yarip.toggleNoFlicker = function(noFlicker, dontNotify, force)
{
    if (noFlicker == this.noFlicker && !force) return false;

    if (noFlicker != true && noFlicker != false) noFlicker = !this.getValue(this.PREF_FLICKER, false, this.DATA_TYPE_BOOLEAN);
    this.noFlicker = noFlicker;

    this.injectCSS("chrome://yarip/skin/noflicker.css", this.noFlicker && this.enabled);
    if (!dontNotify) this.setValue(this.PREF_FLICKER, this.noFlicker, this.DATA_TYPE_BOOLEAN); // notify
    return true;
}

yarip.setMode = function(mode)
{
    var mode = Number(mode);
    if (typeof mode != "number" || mode + "" == "NaN") return false;
    this.mode = mode;
    this.setValue(this.PREF_MODE, mode, this.DATA_TYPE_INTEGER); // notify
    return true;
}

yarip.setAlwaysUseIndex = function(value)
{
    if (value == true || value == false) this.alwaysUseIndex = value;
}

yarip.setElementsInContext = function(value)
{
    var value = Number(value);
    if (typeof value != "number" || value + "" == "NaN") return;
    if (value > 0 && value <= 20) this.elementsInContext = value;
}

yarip.setPurgeInnerHTML = function(value)
{
    if (value == true || value == false) this.purgeInnerHTML = value;
}

yarip.setExclusiveOnCreation = function(value)
{
    if (value == true || value == false) this.exclusiveOnCreation = value;
}

yarip.setTemplates = function(value)
{
    this.templates = [];
    if (!value) return;

    this.templates = value.split(" ", 50); // 50 = max templates
}

yarip.setSchemes = function(value)
{
    try {
        this.schemesRegExp = new RegExp(value);
    } catch (e) {}
}

yarip.setPrivateBrowsing = function(value)
{
    if (value == true || value == false) this.privateBrowsing = value;
}

yarip.getPageRegExp = function(pageName, protocol)
{
    if (!pageName) return "^https?://([^/?#]+\\.)?example\\.net[/?#]";

    protocol = !protocol || /^https?:$/.test(protocol) ? "https?:" : protocol;
//    var domain = pageName.match(this.findDomainRegExp);
//         ^(([^:/?#]+):)?(//([^/?#]*))?([^?#]*)(\?([^#]*))?(#(.*))?
    var domain = pageName.replace(/^(([^:/?#]+):\/+)?|[/?#].*$/g, "").match(this.domainRegExp);
    if (domain) {
        return "^" + protocol + "//" + (!domain[3] /* not IP */ ? "([^/?#]+\\.)?" : "") + domain[0].replace(/([.*+?|()\[\]{}\\])/g, "\\$1") + "[/?#]";
    } else {
        return "^https?://([^/?#]+\\.)?example\\.net[/?#]";
    }
}

yarip.createPage = function(location, pageName, privateBrowsing)
{
    if (!location) return null;

    location = this.getLocationFromLocation(location);
    if (!pageName) pageName = this.getPageName(location, this.mode);

    // Creating the page.
    var page = this.map.get(pageName);
    if (page) {
        page.setTemporary(false);
        return page;
    } else {
        page = new YaripPage(this.getId(), pageName);
        if (privateBrowsing) page.setTemporary(true);
    }

    // Adding the default WhitelistContent.
    if (this.exclusiveOnCreation || privateBrowsing) {
        var sld = location.asciiHost.match(this.findSldRegExp);
        if (sld) {
            var re = this.getPageRegExp(sld[0], location.protocol);
            var content = new YaripContentWhitelistItem(re);
            page.contentWhitelist.add(content);
            page.contentWhitelist.setExclusive(true);
        }
    }

    // Merging with templates.
    var addressObj = this.getAddressObj(pageName, false);
    if (!addressObj.found && this.templates.length > 0) {
        for (var i = 0; i < this.templates.length; i++) {
            page.merge(this.map.get(this.templates[i]));
        }
    }

    // Subtracting duplicate entries.
//    for (var a in addressObj.obj) {
//        page.subtract(this.map.get(a));
//    }

    this.map.add(page);
    this.pageCreated = true;
    return page;
}

yarip.getLocationFromLocation = function(location)
{
    if (!location) return null;
    if (location.isLocation) return location;

    // https://developer.mozilla.org/en/DOM/window.location
    var host = "";
    var asciiHost = "";
//    var hostname = "";
    var port = location.port;
    var asciiHref = "";
    try {
        host = location.host;
        asciiHost = IDNS.convertUTF8toACE(host);
//        hostname = IDNS.convertUTF8toACE(location.hostname);
//        asciiHref = location.protocol + "//" + asciiHost + location.pathname;
        asciiHref = location.protocol + "//" + asciiHost + location.pathname + location.search + location.hash;
//        asciiHref = IDNS.convertUTF8toACE(location.href);
    } catch(e) {}
    return {
//        hash: null,
        host: host,
//        hostname: null,
//        href: protocol + "//" + host + pathname, // no need for search nor hash!
        href: location.href,
        pathname: location.pathname,
        port: port,
        protocol: location.protocol,
//        search: null,
        asciiHost: asciiHost,
//        asciiHref: location.protocol + "//" + asciiHost + pathname,
        asciiHref: asciiHref,
        isLocation: true
    }
}

yarip.getLocationFromContentLocation = function(contentLocation)
{
    if (!contentLocation) return null;
    if (contentLocation.isLocation) return contentLocation;

    var port = "";
//    var host = "";
    var asciiHost = "";
    var hostPort = "";
    try { port = contentLocation.port; } catch (e) {}
//    try { host = contentLocation.host; } catch (e) {}
    try { asciiHost = contentLocation.asciiHost; } catch (e) {}
    try { hostPort = contentLocation.hostPort; } catch (e) {}
    return {
//        hash: null,
        host: hostPort,
//        hostname: null,
        href: contentLocation.spec,
        pathname: contentLocation.path.replace(/[?&#].*$/, ""),
        port: port,
        protocol: contentLocation.scheme + ":",
//        search: null,
//        asciiHost: IDNS.convertUTF8toACE(hostPort),
        asciiHost: asciiHost + (port > -1 && port != 80 ? ":" + port : ""),
        asciiHref: contentLocation.asciiSpec,
        isLocation: true
    }
}

yarip.getContentLocationFromContentLocation = function(contentLocation)
{
    if (!contentLocation) return null;
    if (contentLocation.isContentLocation) return contentLocation;

    // https://developer.mozilla.org/en/nsIURI
    var port = "";
    var host = "";
    var asciiHost = "";
    var hostPort = "";
    try { port = contentLocation.port; } catch (e) {}
    try { host = contentLocation.host; } catch (e) {}
    try { asciiHost = contentLocation.asciiHost; } catch (e) {}
    try { hostPort = contentLocation.hostPort; } catch (e) {}
    return {
        scheme: contentLocation.scheme,
        port: port,
        host: host,
        asciiHost: asciiHost,
        hostPort: hostPort,
        path: contentLocation.path,
        spec: contentLocation.spec,
        asciiSpec: contentLocation.asciiSpec,
        isContentLocation: true
    }
}

yarip.getPageName = function(location, mode)
{
//    if (!location || !/^https?:$/.test(location.protocol)) return null;
    if (!location || !this.schemesRegExp.test(location.protocol.replace(/:$/, ""))) return null;

    var mode = Number(mode);
    if (typeof mode != "number" || mode + "" == "NaN") {
        mode = this.mode;
    }

    location = this.getLocationFromLocation(location);
    var address = null;
    try
    {
        switch (mode) {
        case this.MODE_PAGE:
//            address = location.protocol + "//" + location.host + location.pathname.replace(/[/]*$/, "");
            address = location.protocol + "//" + location.asciiHost + location.pathname.replace(/[/]*$/, "");
//            address = location.asciiHref.replace(/[/]*$/, "");
            break;
        case this.MODE_FQDN:
            address = location.asciiHost;
            break;
        case this.MODE_SLD:
            var sld = null;
            sld = location.asciiHost.match(this.findSldRegExp);
            if (sld) address = sld[0];
            break;
        }
        if (!address) return null;
    } catch (e) {}
    return address;
}

yarip.hasAddress = function(reduceDomain)
{
    return this.getFirstAddress(reduceDomain) != null;
}

yarip.getFirstAddress = function(reduceDomain)
{
    if (!reduceDomain) return null;

    reduceDomain = reduceDomain.replace(/[?&#].*$/, "");
    var tmpDomain = null;
    while (reduceDomain != tmpDomain)
    {
        var reducePath = reduceDomain;
        var tmpPath = null;
        while (reducePath != tmpPath)
        {
            var page = this.map.get(reducePath);
            if (page) return reducePath;

            tmpPath = reducePath;
//            reducePath = reducePath.replace(/((:\d+)?\/|[^/]*)$/, "");
//            reducePath = reducePath.replace(/(\/|[^/]*)$/, "");
            reducePath = reducePath.replace(/(\/|[^/]*)$/, "");
        }
        tmpDomain = reduceDomain;
        reduceDomain = reduceDomain.replace(/^([^:]+:\/\/|[\w-~!$&'()*+,;=:@]+\.)|:\d+$/, "");
    }

    return null;
}

yarip.getAddressMap = function(reduceDomain, follow)
{
    var map = new YaripMap();
    if (typeof reduceDomain == "string") {
        var addressObj = this.getAddressObj(reduceDomain, follow);
        for (var a in addressObj.obj) {
            map.add(this.map.get(a).clone());
        }
    } else { // instanceof Array
        for (var i = 0; i < reduceDomain.length; i++) {
            var addressObj = this.getAddressObj(reduceDomain[i], follow);
            for (var a in addressObj.obj) {
                map.add(this.map.get(a).clone());
            }
        }
    }
    return map;
}

yarip.getAddressObj = function(origReduceDomain, follow, reuse)
{
    if (!origReduceDomain) {
        return {
            obj: {},
            exclusive: false,
            found: false
        };
    }

    var adrObj = reuse && origReduceDomain in this.knownAddressObj ? this.knownAddressObj[origReduceDomain] : null;
    var time = Date.now();
    if (adrObj && time - adrObj.time < 1000) { // < 1 sec
        adrObj.time = time;
        return adrObj;
    }

    var addressObj = {
        obj: {},
        exclusive: false,
        found: false,
        time: time
    }
    var reduceDomain = origReduceDomain.replace(/[?&#].*$/, "");
    var tmpDomain = null;
    while (reduceDomain != tmpDomain)
    {
        var reducePath = reduceDomain;
        var tmpPath = null;
        while (reducePath != tmpPath)
        {
            var page = this.map.get(reducePath);
            if (page) {
                addressObj.obj[reducePath] = [true, true, true, true, true, true, true]; // [isSelf, doElements, doContents, doScripts, doHeaders, doRedirects, doLinks]
                addressObj.found = true;
                if (page.contentWhitelist.getExclusive()) {
                    addressObj.exclusive = true;
                }
            }
            tmpPath = reducePath;
            reducePath = reducePath.replace(/(\/|[^/]*)$/, "");
        }
        tmpDomain = reduceDomain;
        reduceDomain = reduceDomain.replace(/^([^:]+:\/\/|[\w-~!$&'()*+,;=:@]+\.)|:\d+$/, "");
    }

    if (follow) {
        this.getExtensionAddressObj(addressObj, true);
    }

    this.knownAddressObj[origReduceDomain] = addressObj;

    return addressObj;
}

yarip.getExtensionAddressObj = function(addressObj, append)
{
    if (!addressObj) return null;

    if (append) {
        for (var a in addressObj.obj) {
            this.getRecursiveAddressArray(a, addressObj, addressObj.obj[a]);
        }
    } else {
        var aTmp = {
            obj: {},
            exclusive: false,
            found: false
        }
        for (var a in addressObj.obj) {
            this.getRecursiveAddressArray(a, aTmp, addressObj.obj[a]);
        }
        return aTmp;
    }

    return null;
}

yarip.getRecursiveAddressArray = function(parentAddress, addressObj, extMask)
{
    if (!extMask[1] && !extMask[2] && !extMask[3] && !extMask[4] && !extMask[5] && !extMask[6]) return;

    var extObj = this.map.get(parentAddress).pageExtensionList.obj;
    for (var id in extObj)
    {
        var extPage = this.map.getById(id);
        if (extPage)
        {
            var extAddress = extPage.getName();
            var extension = extObj[id];
            if (extAddress in addressObj.obj) { // already added
                if (extMask[1]) addressObj.obj[extAddress][1] = addressObj.obj[extAddress][1] || extension.getDoElements();
                if (extMask[2]) addressObj.obj[extAddress][2] = addressObj.obj[extAddress][2] || extension.getDoContents();
                if (extMask[3]) addressObj.obj[extAddress][3] = addressObj.obj[extAddress][3] || extension.getDoScripts();
                if (extMask[4]) addressObj.obj[extAddress][4] = addressObj.obj[extAddress][4] || extension.getDoHeaders();
                if (extMask[5]) addressObj.obj[extAddress][5] = addressObj.obj[extAddress][5] || extension.getDoRedirects();
                if (extMask[6]) addressObj.obj[extAddress][6] = addressObj.obj[extAddress][6] || extension.getDoLinks();
            } else { // not added
                addressObj.obj[extAddress] = [
                    false, // extended (not self)
                    extMask[1] && extension.getDoElements(),
                    extMask[2] && extension.getDoContents(),
                    extMask[3] && extension.getDoScripts(),
                    extMask[4] && extension.getDoHeaders(),
                    extMask[5] && extension.getDoRedirects(),
                    extMask[6] && extension.getDoLinks()];
                addressObj.found = true;
                if (extPage.pageExtensionList.length > 0) {
                    this.getRecursiveAddressArray(extAddress, addressObj, addressObj.obj[extAddress]);
                }
            }
        } else {
            this.map.get(parentAddress).pageExtensionList.removeByKey(id);
        }
    }
}

yarip.resetUndo = function()
{
    this.undoObj = {};
}

yarip.resetOnAddress = function(obj)
{
    if (!obj) return false;

    var list = null;

    switch (obj.type) {
    case this.TYPE_WHITELIST:
        list = this.map.get(obj.address) ? this.map.get(obj.address).elementWhitelist : null;
        break;
    case this.TYPE_BLACKLIST:
        list = this.map.get(obj.address) ? this.map.get(obj.address).elementBlacklist : null;
        break;
    case this.TYPE_ATTRIBUTE:
        list = this.map.get(obj.address) ? this.map.get(obj.address).elementAttributeList : null;
        break;
    }

    switch (obj.type) {
    case this.TYPE_WHITELIST:
    case this.TYPE_BLACKLIST:
        var elements = this.getElementsByXPath(obj.document, obj.xValue);
        if (elements)
        {
            for (var i = 0; i < elements.snapshotLength; i++)
            {
                var element = elements.snapshotItem(i);
                if (element.nodeType != 1) continue;

                switch (element.getAttribute("status")) {
                case "whitelisted":
                case "whitelisted-forced":
                case "blacklisted":
                    element.removeAttribute("status");
                    break;
                }
            }
        }
        break;
    }

    if (list) {
        list.removeByKey(obj.key);
        this.reloadPage(obj.address, false, true);
    }
    delete this.undoObj[obj.type + "-" + obj.key];
    return true;
}

yarip.getIncrement = function(newIncrement, oldIncrement)
{
    return newIncrement > oldIncrement ? newIncrement : oldIncrement;
}

yarip.reloadPage = function(pageName, selectItem, selectTab)
{
    this.knownAddressObj = {};
    var refresh = this.shouldRefreshPage(pageName);
    if (this.pageDialog && pageName) {
        this.pageDialog.reloadPage(pageName, this.pageCreated || refresh, selectItem, selectTab);
//        this.pageDialog.reloadPage(pageName, this.pageCreated, selectItem, selectTab);
        this.pageCreated = false;
        return true;
    } else {
        return false;
    }
}

yarip.shouldRefreshPage = function(pageName)
{
    if (!pageName) return false;

    var page = this.map.get(pageName);
    if (!page) return false;
//    if (page.isEmpty()) {
//        this.map.remove(page);
//        return true;
//    } else {
//        return false;
//    }
    else return true;
}

yarip.removeAllExceptWhitelisted = function(doc)
{
    this.whitelistXPath(doc, null, "/html/head/descendant-or-self::*[not(@status='blacklisted')]"); // whitelist head and descendants
//    this.blacklistXPath(doc, null, "//*[not(@status='whitelisted')]"); // remove all except whitelisted
    this.blacklistXPath(doc, null, "//*[not(@status='whitelisted') and not(@status='whitelisted-forced')]"); // remove all except whitelisted
}

//yarip.whitelistXPath = function(doc, pageName, xValue, isNew, xpath, doIncrement)
yarip.whitelistXPath = function(doc, pageName, xValue, isNew, force, xpath, doIncrement)
{
//    if (!doc || !pageName || !xValue) return false;
    if (!doc || !xValue) return false;

    var elements = this.getElementsByXPath(doc, xValue);
    if (!elements) return false;

//    var page = this.map.get(pageName);
//    var elementWhitelist = page ? page.elementWhitelist : null;
    var found = false;
    var increment = this.INCREMENT_NOT_FOUND;

    for (var i = 0; i < elements.snapshotLength; i++)
    {
        var element = elements.snapshotItem(i);

        switch (element.nodeType) {
        case 1: // ELEMENT
            switch (element.getAttribute("status")) {
            case "whitelisted":
            case "whitelisted-forced":
            case "blacklisted":
                increment = this.getIncrement(this.INCREMENT_IGNORED, increment);
                continue;
            }

            if (isNew) {
                element.setAttribute("status", force ? "whitelisted-forced" : "whitelisted");
            } else {
                while (element && element.nodeType == 1 && element.getAttribute("status") != "whitelisted" && element.getAttribute("status") != "whitelisted-forced") {
                    element.setAttribute("status", force ? "whitelisted-forced" : "whitelisted");
                    element = element.parentNode;
                }
            }
            break;

        default:
            continue;
        }

        increment = this.INCREMENT_FOUND;
        found = true;
    }

    if (!pageName) return found;

    var page = this.map.get(pageName);
    var elementWhitelist = page ? page.elementWhitelist : null;

    if (doIncrement && elementWhitelist && elementWhitelist.obj[xValue])
    {
        switch (increment) {
        case this.INCREMENT_FOUND:
            elementWhitelist.obj[xValue].incrementFound();
            break;
        case this.INCREMENT_NOT_FOUND:
            elementWhitelist.obj[xValue].incrementNotFound();
            break;
        case this.INCREMENT_IGNORED:
            elementWhitelist.obj[xValue].incrementIgnored();
            break;
        }
    }

    if (isNew && found)
    {
        var xpath = xpath ? xpath : new YaripElementWhitelistItem(xValue);
        if (page) {
            page.setTemporary(false);
        } else {
            page = this.createPage(doc.location, pageName);
        }
        page.elementWhitelist.add(xpath);
        this.reloadPage(pageName, false, true);

        var obj = {};
        obj.document = doc;
        obj.address = pageName;
        obj.xValue = xValue;
        obj.key = xpath.getKey();
        obj.type = this.TYPE_WHITELIST;
        obj.text = xValue;
        this.undoObj[obj.type + "-" + obj.key] = obj;
    }

    return found;
}

yarip.blacklistXPath = function(doc, pageName, xValue, isNew, force, xpath, doIncrement)
{
//    if (!doc || !pageName || !xValue) return false;
    if (!doc || !xValue) return false;

    var elements = this.getElementsByXPath(doc, xValue);
    if (!elements) return false;

    var found = false;
    var increment = this.INCREMENT_NOT_FOUND;
    var useForce = true;

    for (var i = 0; i < elements.snapshotLength; i++)
    {
        var element = elements.snapshotItem(i);
        switch (element.nodeType) {
        case 1: // ELEMENT
            switch (element.getAttribute("status")) {
            case "whitelisted":
                if (force) {
                  useForce = false;
                  break;
                } else {
                  increment = this.getIncrement(this.INCREMENT_IGNORED, increment);
                  continue;
                }
            case "whitelisted-forced":
              increment = this.getIncrement(this.INCREMENT_IGNORED, increment);
              continue;
            case "blacklisted":
                increment = this.getIncrement(this.INCREMENT_IGNORED, increment);
                continue;
            }

            switch (element.localName) {
            case "html":
            case "body":
            case "frameset":
                continue;
            case "head":
                element.innerHTML =
                    "<style rel='stylesheet' type='text/css'>\n" +
                        "\t* {\n" +
                            "\t\tmargin: 0 !important;\n" +
                            "\t\tpadding: 0 !important;\n" +
                        "\t}\n" +
                        "\tdiv {\n" +
                            "\t\theight: auto !important;\n" +
                        "\t}\n" +
                        "\ttable {\n" +
                            "\t\tborder-spacing: 0 0 !important;\n" +
                            "\t\theight: auto !important;\n" +
                            "\t\twidth: 0 !important;\n" +
                        "\t}\n" +
                    "</style>";
                break;
            case "script":
                element.removeAttribute("src");
                element.innerHTML = "";
                break;
            case "frame":
                if (element.parentNode.localName == "FRAMESET") {
                    continue;
                }
                element.setAttribute("src", "about:blank");
                element.setAttribute("status", "blacklisted");
                break;
//            case "embed":
//            case "iframe":
//                element.setAttribute("src", "about:blank");
//                element.setAttribute("status", "blacklisted");
//                break;
            case "br":
            case "embed":
            case "iframe":
                element.setAttribute("status", "blacklisted");
                break;
            default:
                element.setAttribute("status", "blacklisted");
                if (this.purgeInnerHTML && !isNew) {
                    element.innerHTML = "";
                }
                break;
            }
            break;
        case 2: // ATTRIBUTE
            if (element.ownerElement) element.ownerElement.removeAttribute(element.name);
            break;
        case 3: // TEXT
            if (element.parentNode) element.parentNode.removeChild(element);
            break;
        default:
            continue;
        }

        increment = this.INCREMENT_FOUND;
        found = true;
    }

    if (!pageName) return found;

    var page = this.map.get(pageName);
    var elementBlacklist = page ? page.elementBlacklist : null;

    if (doIncrement && elementBlacklist && elementBlacklist.obj[xValue])
    {
        switch (increment) {
        case this.INCREMENT_FOUND:
            elementBlacklist.obj[xValue].incrementFound();
            break;
        case this.INCREMENT_NOT_FOUND:
            elementBlacklist.obj[xValue].incrementNotFound();
            break;
        case this.INCREMENT_IGNORED:
            elementBlacklist.obj[xValue].incrementIgnored();
            break;
        }
    }

    if (isNew && found)
    {
        var xpath = xpath ? xpath : new YaripElementBlacklistItem(xValue, true, useForce);
        if (page) {
            page.setTemporary(false);
        } else {
            page = this.createPage(doc.location, pageName);
        }
        page.elementBlacklist.add(xpath);
        this.reloadPage(pageName, false, true);

        var obj = {};
        obj.document = doc;
        obj.address = pageName;
        obj.xValue = xValue;
        obj.key = xpath.getKey();
        obj.type = this.TYPE_BLACKLIST;
        obj.text = xValue;
        this.undoObj[obj.type + "-" + obj.key] = obj;
    }

    return found;
}

yarip.whitelistContent = function(pageName, regExpValue, content, location)
{
    if (!pageName || !regExpValue) return false;

    var tmpContent = content ? content : new YaripContentWhitelistItem(regExpValue);
    var page = this.map.get(pageName);
    if (page) {
        page.setTemporary(false);
    } else {
        page = this.createPage(location, pageName);
    }
    page.contentWhitelist.add(tmpContent);
    this.reloadPage(pageName, false, true);
    return true;
}

yarip.blacklistContent = function(pageName, regExpValue, content, location)
{
    if (!pageName || !regExpValue) return false;

    var tmpContent = content ? content : new YaripContentBlacklistItem(regExpValue);
    var page = this.map.get(pageName);
    if (page) {
        page.setTemporary(false);
    } else {
        page = this.createPage(location, pageName);
    }
    page.contentBlacklist.add(tmpContent);
    this.reloadPage(pageName, false, true);
    return true;
}

//yarip.extendPage = function(pageName, contentLocation, contentAddress)
yarip.extendPage = function(pageLocation, pageName, contentLocation, contentAddress)
{
    if (!pageName || !contentAddress) return false;

//    var pageOwn = this.map.get(pageName);
//    if (!pageOwn) return false;
    var pageOwn = this.createPage(pageLocation, pageName);
    var pageExt = this.createPage(contentLocation, contentAddress);
    pageOwn.pageExtensionList.add(new YaripPageExtensionItem(pageExt.getId(), false, true, false, false, false, false));
    pageOwn.pageExtensionList.removeByKey(pageOwn.getId());
    pageOwn.setTemporary(false);
    this.reloadPage(pageExt.getName()); // Reloaded if page created!
    this.reloadPage(pageOwn.getName()); // Reloaded if page created!
    return true;
}

yarip.styleElement = function(doc, pageName, xValue, attrName, attrValue, isNew, xpath, doIncrement)
{
    if (!doc || !pageName || !xValue || !attrName || !attrValue) return false;

    var elements = this.getElementsByXPath(doc, xValue);
    if (!elements) return false;

    var found = false;
    var increment = this.INCREMENT_NOT_FOUND;

    for (var i = 0; i < elements.snapshotLength; i++)
    {
        var element = elements.snapshotItem(i);
        switch (element.nodeType) {
        case 1: // ELEMENT
            switch (attrName) {
            case "style":
                var style = element.getAttribute(attrName);
                element.setAttribute(attrName, style ? style + attrValue : attrValue);
                break;
            default:
                element.setAttribute(attrName, attrValue);
                break;
            }
            break;
        default:
            continue;
        }

        increment = this.INCREMENT_FOUND;
        found = true;
    }

    var page = this.map.get(pageName);
    var list = page ? page.elementAttributeList : null;

    xpath = xpath ? xpath : new YaripElementAttributeItem(xValue, attrName, attrValue);
    var key = xpath.getKey();
    if (doIncrement && list && list.obj[key])
    {
        switch (increment) {
        case this.INCREMENT_FOUND:
            list.obj[key].incrementFound();
            break;
        case this.INCREMENT_NOT_FOUND:
            list.obj[key].incrementNotFound();
            break;
        }
    }

    if (isNew && found)
    {
        if (page) {
            page.setTemporary(false);
        } else {
            page = this.createPage(doc.location, pageName);
        }
        page.elementAttributeList.add(xpath);
        this.reloadPage(pageName, false, true);

        var obj = {};
        obj.document = doc;
        obj.address = pageName;
        obj.key = xpath.getKey();
        obj.type = this.TYPE_ATTRIBUTE;
        obj.text = attrName + ": " + attrValue;
        this.undoObj[obj.type + "-" + obj.key] = obj;
    }

    return found;
}

yarip.scriptElement = function(doc, pageName, xValue, isNew, xpath)
{
    if (!doc || !pageName || !xValue || !xpath) return false;

    var elements = this.getElementsByXPath(doc, xValue);
    if (!elements) return false;

    var page = this.map.get(pageName);

    if (isNew)
    {
//        xpath = xpath ? xpath : new YaripElementScriptItem(xValue,
//                    "function (array) {\n" +
//                    "    // Do something with the elements in the array!\n" +
//                    "}");
        if (page) {
            page.setTemporary(false);
        } else {
            page = this.createPage(doc.location, pageName);
        }
        page.elementScriptList.add(xpath);
        this.reloadPage(pageName, false, true);
    }

    return true;
}

yarip.calculateXPath = function(element, fullPath)
{
    if (!element || !element.localName) return null;

    if (element.nodeType == 2) {
        return this.calculateXPath(element.ownerElement, fullPath) + "/@" + element.localName; // attribute
    }

    var xValue = "";
    do
    {
        var attributes = "";
        if (element.getAttribute("id"))
        {
            // ID
            attributes += "[@id='" + element.getAttribute("id") + "']";
            if (!fullPath) {
                return "//" + element.localName + attributes + xValue; // shortest xpath
            }
        }
        else
        {
            // ATTRIBUTES
            if (element.getAttribute("name")) attributes += "@name='" + element.getAttribute("name") + "' and ";
            if (element.getAttribute("class")) attributes += "@class='" + element.getAttribute("class") + "' and ";
            if (attributes != "") attributes = "[" + attributes.substring(0, attributes.length - 5) + "]";

            // INDEX
            if (!/^(html|head|title|body|thead|tbody|tfoot|frameset)$/i.test(element.localName))
            {
                var index = 1;
                var sibling = element.previousSibling;
                while (sibling) {
                    if (sibling.localName == element.localName) ++index;
                    sibling = sibling.previousSibling;
                }

                var found = false;
                sibling = element.nextSibling;
                while (!found && sibling) {
                    if (sibling.localName == element.localName) found = true;
                    else sibling = sibling.nextSibling;
                }
                if (found || index > 1 || this.alwaysUseIndex) attributes = "[" + index + "]" + attributes;
            }
        }
        xValue = "/" + element.localName + attributes + xValue;
        element = element.parentNode;
    } while (element && element.nodeType == 1);

    return xValue != "" ? xValue : null;
}

yarip.getElementsByXPath = function(doc, xValue)
{
    var elements = null;
    try {
        elements = doc.evaluate(xValue, doc.body, null, Ci.nsIDOMXPathResult.UNORDERED_NODE_SNAPSHOT_TYPE, null);
    } catch (e) {
        var consoleService = Cc['@mozilla.org/consoleservice;1'].getService(Ci.nsIConsoleService);
        var scriptError = Cc['@mozilla.org/scripterror;1'].createInstance(Ci.nsIScriptError);
//        var message = doc.getElementById("yarip-overlay-stringbundle").getFormattedString("malformedXPath", [xValue]);
        var message = "[xpath=" + xValue + "] ";
//        var sourceName = "chrome://yarip/content/yarip.js";
        var sourceName = "";
        scriptError.init(message + e.message, sourceName, null, e.lineNumber, e.columnNumber, scriptError.errorFlag, "content javascript");
        consoleService.logMessage(scriptError);
    }
//    if (!elements || elements.snapshotLength == 0) return null;
    return elements;
}

yarip.generateRegExp = function(value)
{
    if (!value) return null;
//    return "^" + value.replace(/([/.*+?|()\[\]{}\\])/g, "\\$1") + "$";
    return "^" + value.replace(/([.*+?|()\[\]{}\\])/g, "\\$1") + "$";
}

yarip.checkPageName = function(value)
{
    return this.uriSimpleRegExp.test(value);
}

yarip.checkXPath = function(xValue)
{
    if (!xValue) return false;
    try
    {
        var dp = Cc['@mozilla.org/xmlextras/domparser;1'].createInstance(Ci.nsIDOMParser);
        var doc = dp.parseFromString("<yarip></yarip>", "text/xml");
        doc.evaluate(xValue, doc, null, Ci.nsIDOMXPathResult.UNORDERED_NODE_SNAPSHOT_TYPE, null);
    }
    catch (e) { return false; }
    return true;
}

yarip.checkRegExp = function(reValue)
{
    if (!reValue) return false;
    try { new RegExp(reValue).test(""); } catch (e) { return false; }
    return true;
}

yarip.getElements = function(doc, xValue)
{
    var elements = null;
    try { elements = doc.evaluate(xValue, doc, null, Ci.nsIDOMXPathResult.UNORDERED_NODE_SNAPSHOT_TYPE, null); } catch (e) {}
    return elements;
}

yarip.highlight = function(doc, xValue)
{
    if (!doc || !xValue) return false;
    var elements = this.getElements(doc, xValue);
    if (!elements) return false;
    if (elements.snapshotLength > 0) {
        for (var i = 0; i < elements.snapshotLength; i++) {
            var element = elements.snapshotItem(i);
            if (element && element.getAttribute) {
                var status = element.getAttribute("status");
                if (!status) {
                    element.setAttribute("status", "highlighted");
                } else if (!/\bhighlighted\b/.test(status)) {
                    element.setAttribute("status", status + " highlighted");
                }
            }
        }
    }
    return true;
}

yarip.unHighlight = function(doc, xValue)
{
    if (!doc || !xValue) return;
    var elements = this.getElements(doc, xValue);
    if (elements && elements.snapshotLength > 0) {
        for (var i = 0; i < elements.snapshotLength; i++) {
            var element = elements.snapshotItem(i);
            if (element && element.getAttribute) {
                var status = element.getAttribute("status");
                if (status && /\bhighlighted\b/.test(status)) {
                    var newStatus = status.replace(/\s*highlighted\b/, "");
//                    if (newStatus == "") {
                    if (!newStatus) {
                        element.removeAttribute("status");
                    } else {
                        element.setAttribute("status", newStatus);
                    }
                }
            }

        }
    }
}

yarip.save = function()
{
    var file = this.getFile(this.FILE);
    var data = this.map.generateXml();
    this.saveToFile(data, file);
}

yarip.getValue = function(preference, defaultValue, type)
{
    try {
        switch (type) {
        case this.DATA_TYPE_BOOLEAN:
            return PB.getBoolPref(preference);
        case this.DATA_TYPE_INTEGER:
            return PB.getIntPref(preference);
        case this.DATA_TYPE_STRING:
            return PB.getCharPref(preference);
        }
    } catch (e) {
        if (defaultValue != false && (typeof defaultValue != "number" || defaultValue + "" == "NaN") && !defaultValue) {
            return null;
        }
        this.setValue(preference, defaultValue, type);
    }
    return this.getValue(preference, null, type);
}

yarip.setValue = function(preference, value, type)
{
    try {
        switch (type) {
        case this.DATA_TYPE_BOOLEAN:
            PB.setBoolPref(preference, "" + value == "true");
            break;
        case this.DATA_TYPE_INTEGER:
            PB.setIntPref(preference, value);
            break;
        case this.DATA_TYPE_STRING:
            PB.setCharPref(preference, value);
            break;
        case this.DATA_TYPE_RESET:
            PB.clearUserPref(preference);
            break;
        }
    } catch (e) {}
}

yarip.deleteBranch = function(branch)
{
    if (!branch) return;

    PB.deleteBranch(branch);
}

yarip.injectCSS = function(uri, value)
{
    if (!uri) return;

    var uri = IOS.newURI(uri, this.CHARSET, null); // uri, charset, baseuri
    var sss = Cc["@mozilla.org/content/style-sheet-service;1"].getService(Ci.nsIStyleSheetService);
    if (value) { if (!sss.sheetRegistered(uri, sss.USER_SHEET)) sss.loadAndRegisterSheet(uri, sss.USER_SHEET); }
    else { if (sss.sheetRegistered(uri, sss.USER_SHEET)) sss.unregisterSheet(uri, sss.USER_SHEET); }
}

yarip.getCurrentVersion = function()
{
    return this.VERSION;
}

yarip.check = function(file)
{
    var previous = this.getValue(this.PREF_VERSION, "", this.DATA_TYPE_STRING);
    var current = this.getCurrentVersion();
    if (!previous && previous != current)
    {
        this.setValue(this.PREF_VERSION, current, this.DATA_TYPE_STRING);

        if (/^0\.(1\.[2-9]|2(\.1)?)$/.test(previous) && !/^0\.(1\.[2-9]|2(\.1)?)$/.test(current)) {
            // CHANGED NOFLICKER FEATURE
            this.setValue("content.notify.ontimer", null, this.DATA_TYPE_RESET);
            this.setValue("content.notify.backoffcount", null, this.DATA_TYPE_RESET);
            this.setValue("nglayout.initialpaint.delay", null, this.DATA_TYPE_RESET);
        }

        if (/^0\.2(\.[1-3])?$/.test(previous) && !/^0\.2(\.[1-3])?$/.test(current)) {
            Cc["@mozilla.org/satchel/form-history;1"].getService(Ci.nsIFormHistory2).removeEntriesForName("style");
        }

        if (/^0\.2\.4$/.test(previous) && !/^0\.2\.4$/.test(current)) {
            this.setSchemes(this.getValue("extensions.yarip.schemesList.value", "^https?$", this.DATA_TYPE_STRING));
            this.deleteBranch("extensions.yarip.schemesList");
        }

        this.deleteBranch("extensions.yarip.changeType"); // >0.1.7
        this.deleteBranch("extensions.yarip.flickerType"); // >0.1.7
        this.deleteBranch("extensions.yarip.loggingEnabled"); // >0.2.1
        this.deleteBranch("extensions.yarip.deep"); // >0.2.2
        this.deleteBranch("extensions.yarip.profile"); // >0.2.2
        this.deleteBranch("extensions.yarip.reload"); // >0.2.2
        this.deleteBranch("extensions.yarip.firstrun"); // >0.2.4

        if (/^0\.(1\.[2-9]|2(\.[1-4])?)$/.test(previous) && !/^0\.(1\.[2-9]|2(\.[1-4])?)$/.test(current)) {
            // TODO Remove the STYLE-file!
            //      -> "extensions/yarip@mozdev.org-dev/style.css"
        }
    }
}

yarip.getId = function()
{
    var id = Date.now();
    while (this.map.getById(id) != undefined && this.map.getById(id) != null) id++;
    return "" + id;
}

yarip.addMonitorDialog = function(monitorDialog)
{
    monitorDialog.id = Date.now();
    while (this.monitorDialogues["" + monitorDialog.id] != undefined) monitorDialog.id++;
    this.monitorDialogues["" + monitorDialog.id] = monitorDialog;
}

yarip.removeMonitorDialog = function(monitorDialog)
{
    delete this.monitorDialogues["" + monitorDialog.id];
    monitorDialog.id = -1;
}

yarip.load = function(file, imported)
{
    var doc = this.getDoc(file);
    if (!doc) {
        if (!imported) this.save();
        return null;
    } else if (doc.documentElement.nodeName == "parsererror") {
        return null;
    }

    var pageName = null;
    var map = new YaripMap();

    var iYarip = doc.evaluate("./yarip", doc, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
    var nYarip = iYarip.iterateNext();
    if (!nYarip) return null;

    var version = nYarip.getAttribute("version");
    if (!version)
    {
        var iPage = doc.evaluate("./page", nYarip, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
        var nPage = iPage.iterateNext();
        while (nPage)
        {
            pageName = nPage.getAttribute("name");
            var pageId = nPage.getAttribute("id");
            var page = map.obj[pageName] ? map.obj[pageName] : new YaripPage(pageId ? pageId : this.getId(), pageName);

            // XPATH WHITELIST
            var iWhitelist = doc.evaluate("./whitelist", nPage, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
            var nWhitelist = iWhitelist.iterateNext();
            if (nWhitelist)
            {
                var iXPath = doc.evaluate("./xpath", nWhitelist, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                var nXPath = iXPath.iterateNext();
                while (nXPath)
                {
                    page.elementWhitelist.add(new YaripElementWhitelistItem(
                        nXPath.getAttribute("value"),
                        nXPath.getAttribute("deep"),
                        null, // force
                        pageId, // created
                        null, // lastFound
                        nXPath.getAttribute("found"),
                        nXPath.getAttribute("notFound"),
                        nXPath.getAttribute("whitelisted")));
                    nXPath = iXPath.iterateNext();
                }
            }

            // XPATH BLACKLIST
            var iBlacklist = doc.evaluate("./blacklist", nPage, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
            var nBlacklist = iBlacklist.iterateNext();
            if (nBlacklist)
            {
                var iXPath = doc.evaluate("./xpath", nBlacklist, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                var nXPath = iXPath.iterateNext();
                while (nXPath)
                {
                    page.elementBlacklist.add(new YaripElementBlacklistItem(
                        nXPath.getAttribute("value"),
                        nXPath.getAttribute("deep"),
                        nXPath.getAttribute("force") == "true",
                        nXPath.getAttribute("dependsOnWhitelist"),
                        pageId, // created
                        null, // lastFound
                        nXPath.getAttribute("found"),
                        nXPath.getAttribute("notFound"),
                        nXPath.getAttribute("whitelisted")));
                    nXPath = iXPath.iterateNext();
                }
            }

            map.add(page);
            nPage = iPage.iterateNext();
        }
    }
    else if (/^0\.2(\.[1-3])?$/.test(version))
    {
        var iPage = doc.evaluate("./page", nYarip, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
        var nPage = iPage.iterateNext();
        while (nPage)
        {
            pageName = nPage.getAttribute("name");
            var pageId = nPage.getAttribute("id");
            var page = map.obj[pageName] ? map.obj[pageName] : new YaripPage(pageId ? pageId : this.getId(), pageName);

            // ELEMENT BLACKLIST
            var iElementBlacklist = doc.evaluate("./elementBlacklist", nPage, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
            var nElementBlacklist = iElementBlacklist.iterateNext();
            if (nElementBlacklist)
            {
                var iElement = doc.evaluate("./element", nElementBlacklist, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                var nElement = iElement.iterateNext();
                while (nElement)
                {
                    var element = new YaripElementBlacklistItem(
                        nElement.getAttribute("xpath"),
                        nElement.getAttribute("deep"),
                        nElement.getAttribute("force") == "true",
                        nElement.getAttribute("dow"),
                        pageId, // created
                        null, // lastFound
                        nElement.getAttribute("found"),
                        nElement.getAttribute("notFound"),
                        nElement.getAttribute("ignored"));
                    page.elementBlacklist.add(element);
                    nElement = iElement.iterateNext();
                }
            }

            // CONTENT BLACKLIST
            var iContentBlacklist = doc.evaluate("./contentBlacklist", nPage, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
            var nContent = iContentBlacklist.iterateNext();
            if (nContent)
            {
                var iContent = doc.evaluate("./content", nContent, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                var nContent = iContent.iterateNext();
                while (nContent)
                {
                    page.contentBlacklist.add(new YaripContentBlacklistItem(
                        nContent.getAttribute("regexp"),
                        nContent.getAttribute("deep"),
                        nContent.getAttribute("force") == "true",
                        pageId, // created
                        null, // lastFound
                        nContent.getAttribute("ignored")));
                    nContent = iContent.iterateNext();
                }
            }

            map.add(page);
            nPage = iPage.iterateNext();
        }

        var iPage = doc.evaluate("./page", nYarip, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
        var nPage = iPage.iterateNext();
        while (nPage)
        {
            pageName = nPage.getAttribute("name");
            var pageId = nPage.getAttribute("id");
            var page = map.obj[pageName] ? map.obj[pageName] : new YaripPage(pageId ? pageId : this.getId(), pageName);

            // ELEMENT WHITELIST
            var iElementWhitelist = doc.evaluate("./elementWhitelist", nPage, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
            var nElementWhitelist = iElementWhitelist.iterateNext();
            if (nElementWhitelist)
            {
//                page.elementWhitelist.setExclusive(nElementWhitelist.getAttribute("exclusive"));
                var iElement = doc.evaluate("./element", nElementWhitelist, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                var nElement = iElement.iterateNext();
                while (nElement)
                {
                    page.elementWhitelist.add(new YaripElementWhitelistItem(
                        nElement.getAttribute("xpath"),
                        nElement.getAttribute("deep"),
                        null, // force
                        pageId, // created
                        null, // lastFound
                        nElement.getAttribute("found"),
                        nElement.getAttribute("notFound"),
                        nElement.getAttribute("ignored")));
                    nElement = iElement.iterateNext();
                }
            }

            // CONTENT WHITELIST
            var iContentWhitelist = doc.evaluate("./contentWhitelist", nPage, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
            var nContent = iContentWhitelist.iterateNext();
            if (nContent)
            {
                page.contentWhitelist.setExclusive(nContent.getAttribute("exclusive"));
                var iContent = doc.evaluate("./content", nContent, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                var nContent = iContent.iterateNext();
                while (nContent)
                {
                    page.contentWhitelist.add(new YaripContentWhitelistItem(
                        nContent.getAttribute("regexp"),
                        nContent.getAttribute("deep"),
                        null, // force
                        pageId, // created
                        null)); // lastFound
                    nContent = iContent.iterateNext();
                }
            }

            // STYLES
            var iStyles = doc.evaluate("./styles", nPage, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
            var nStyles = iStyles.iterateNext();
            if (nStyles)
            {
                var iElement = doc.evaluate("./element", nStyles, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                var nElement = iElement.iterateNext();
                while (nElement)
                {
                    page.elementAttributeList.add(new YaripElementAttributeItem(
                        nElement.getAttribute("xpath"),
                        "style",
                        nElement.getAttribute("style"),
                        nElement.getAttribute("deep"),
                        pageId, // created
                        null, // lastFound
                        nElement.getAttribute("found"),
                        nElement.getAttribute("notFound")));
                    nElement = iElement.iterateNext();
                }
            }

            // TODO Remove this, since it was a temporary change!
            // ATTRIBUTES
            var iAttributes = doc.evaluate("./attributes", nPage, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
            var nAttributes = iAttributes.iterateNext();
            if (nAttributes)
            {
                var iElement = doc.evaluate("./element", nAttributes, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                var nElement = iElement.iterateNext();
                while (nElement)
                {
                    page.elementAttributeList.add(new YaripElementAttributeItem(
                        nElement.getAttribute("xpath"),
                        nElement.getAttribute("name"),
                        nElement.getAttribute("value"),
                        nElement.getAttribute("deep"),
                        pageId, // created
                        nElement.getAttribute("lastFound"),
                        nElement.getAttribute("found"),
                        nElement.getAttribute("notFound")));
                    nElement = iElement.iterateNext();
                }
            }

            map.add(page);
            nPage = iPage.iterateNext();
        }

        if (imported) {
            this.map.merge(map);
        } else {
            this.map = map;
        }

        var iPage = doc.evaluate("./page", nYarip, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
        var nPage = iPage.iterateNext();
        while (nPage)
        {
            pageName = nPage.getAttribute("name");
            var pageId = nPage.getAttribute("id");
            var page = map.obj[pageName];

            // EXTENSION
            var iExtension = doc.evaluate("./extension", nPage, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
            var nExtension = iExtension.iterateNext();
            if (nExtension)
            {
                var iItem = doc.evaluate("./item", nExtension, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                var nItem = iItem.iterateNext();
                while (nItem)
                {
                    var value = nItem.getAttribute("value");
                    if (value != page.getId()) {
                        page.pageExtensionList.add(new YaripPageExtensionItem(
                            nItem.getAttribute("value"),
                            nItem.getAttribute("doElements"),
                            nItem.getAttribute("doContents"),
                            nItem.getAttribute("doScripts"),
                            false,
                            false,
                            false,
                            pageId)); // created
                    }
                    nItem = iItem.iterateNext();
                }
                page.pageExtensionList.removeByKey(page.getId()); // remove self
            }

            nPage = iPage.iterateNext();
        }
    }
    else if (/^0\.2\.4$/.test(version))
    {
        var iPage = doc.evaluate("./page", nYarip, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
        var nPage = iPage.iterateNext();
        while (nPage)
        {
            pageName = nPage.getAttribute("name");
            var pageId = nPage.getAttribute("id");
            var page = map.obj[pageName] ? map.obj[pageName] : new YaripPage(pageId ? pageId : this.getId(), pageName);

            // ELEMENT WHITELIST
            var iElementWhitelist = doc.evaluate("./elementWhitelist", nPage, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
            var nElementWhitelist = iElementWhitelist.iterateNext();
            if (nElementWhitelist)
            {
                var iElement = doc.evaluate("./element", nElementWhitelist, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                var nElement = iElement.iterateNext();
                while (nElement)
                {
                    var iXPath = doc.evaluate("./xpath", nElement, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                    var nXPath = iXPath.iterateNext();
                    if (nXPath) {
                        var created = nElement.getAttribute("created");
                        page.elementWhitelist.add(new YaripElementWhitelistItem(
                            nXPath.textContent,
                            nElement.getAttribute("deep"),
                            null, // force
                            created ? created : pageId,
                            nElement.getAttribute("lastFound"),
                            nElement.getAttribute("found"),
                            nElement.getAttribute("notFound"),
                            nElement.getAttribute("ignored")));
                    }
                    nElement = iElement.iterateNext();
                }
            }

            // ELEMENT BLACKLIST
            var iElementBlacklist = doc.evaluate("./elementBlacklist", nPage, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
            var nElementBlacklist = iElementBlacklist.iterateNext();
            if (nElementBlacklist)
            {
                var iElement = doc.evaluate("./element", nElementBlacklist, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                var nElement = iElement.iterateNext();
                while (nElement)
                {
                    var iXPath = doc.evaluate("./xpath", nElement, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                    var nXPath = iXPath.iterateNext();
                    if (nXPath) {
                        var created = nElement.getAttribute("created");
                        page.elementBlacklist.add(new YaripElementBlacklistItem(
                            nXPath.textContent,
                            nElement.getAttribute("deep"),
                            nElement.getAttribute("force"),
                            nElement.getAttribute("dow"),
                            created ? created : pageId,
                            nElement.getAttribute("lastFound"),
                            nElement.getAttribute("found"),
                            nElement.getAttribute("notFound"),
                            nElement.getAttribute("ignored")));
                    }
                    nElement = iElement.iterateNext();
                }
            }

            // CONTENT WHITELIST
            var iContentWhitelist = doc.evaluate("./contentWhitelist", nPage, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
            var nContent = iContentWhitelist.iterateNext();
            if (nContent)
            {
                page.contentWhitelist.setExclusive(nContent.getAttribute("exclusive"));
                var iContent = doc.evaluate("./content", nContent, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                var nContent = iContent.iterateNext();
                while (nContent)
                {
                    var iRegexp = doc.evaluate("./regexp", nContent, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                    var nRegexp = iRegexp.iterateNext();
                    if (nRegexp) {
                        var created = nContent.getAttribute("created");
                        page.contentWhitelist.add(new YaripContentWhitelistItem(
                            nRegexp.textContent,
                            nContent.getAttribute("deep"),
                            null,
                            created ? created : pageId,
                            nContent.getAttribute("lastFound")));
                    }
                    nContent = iContent.iterateNext();
                }
            }

            // CONTENT BLACKLIST
            var iContentBlacklist = doc.evaluate("./contentBlacklist", nPage, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
            var nContent = iContentBlacklist.iterateNext();
            if (nContent)
            {
                var iContent = doc.evaluate("./content", nContent, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                var nContent = iContent.iterateNext();
                while (nContent)
                {
                    var iRegexp = doc.evaluate("./regexp", nContent, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                    var nRegexp = iRegexp.iterateNext();
                    if (nRegexp) {
                        var created = nContent.getAttribute("created");
                        page.contentBlacklist.add(new YaripContentBlacklistItem(
                            nRegexp.textContent,
                            nContent.getAttribute("deep"),
                            nContent.getAttribute("force"),
                            created ? created : pageId,
                            nContent.getAttribute("lastFound"),
                            nContent.getAttribute("ignored")));
                    }
                    nContent = iContent.iterateNext();
                }
            }

            // ATTRIBUTES
            var iAttributes = doc.evaluate("./attributes", nPage, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
            var nAttributes = iAttributes.iterateNext();
            if (nAttributes)
            {
                var iElement = doc.evaluate("./element", nAttributes, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                var nElement = iElement.iterateNext();
                while (nElement)
                {
                    var iXPath = doc.evaluate("./xpath", nElement, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                    var nXPath = iXPath.iterateNext();
                    var iName = doc.evaluate("./name", nElement, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                    var nName = iName.iterateNext();
                    var iValue = doc.evaluate("./value", nElement, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                    var nValue = iValue.iterateNext();
                    if (nXPath && nName && nValue) {
                        var created = nElement.getAttribute("created");
                        page.elementAttributeList.add(new YaripElementAttributeItem(
                            nXPath.textContent,
                            nName.textContent,
                            nValue.textContent,
                            nElement.getAttribute("deep"),
                            created ? created : pageId,
                            nElement.getAttribute("lastFound"),
                            nElement.getAttribute("found"),
                            nElement.getAttribute("notFound")));
                    }
                    nElement = iElement.iterateNext();
                }
            }

            // SCRIPT
            var iScript = doc.evaluate("./script", nPage, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
            var nScript = iScript.iterateNext();
            if (nScript) {
                page.pageScriptList.setLastChanged(nScript.getAttribute("lastChanged"));
                page.pageScriptList.set(null, null, nScript.textContent);
            }

            map.add(page);
            nPage = iPage.iterateNext();
        }

        if (imported) {
            this.map.merge(map);
        } else {
            this.map = map;
        }

        var iPage = doc.evaluate("./page", nYarip, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
        var nPage = iPage.iterateNext();
        while (nPage)
        {
            pageName = nPage.getAttribute("name");
            var pageId = nPage.getAttribute("id");
            var page = map.obj[pageName];

            // EXTENSION
            var iExtension = doc.evaluate("./extension", nPage, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
            var nExtension = iExtension.iterateNext();
            if (nExtension)
            {
                var iPageRef = doc.evaluate("./pageReference", nExtension, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                var nPageRef = iPageRef.iterateNext();
                while (nPageRef)
                {
                    var pageExtId = nPageRef.textContent;
                    if (pageExtId && pageExtId != page.getId()) {
                        var created = nPageRef.getAttribute("created");
                        page.pageExtensionList.add(new YaripPageExtensionItem(
                            pageExtId,
                            nPageRef.getAttribute("doElements"),
                            nPageRef.getAttribute("doContents"),
                            nPageRef.getAttribute("doScripts"),
                            nPageRef.getAttribute("doHeaders"),
                            nPageRef.getAttribute("doRedirects"),
                            nPageRef.getAttribute("doLinks"),
                            created ? created : pageId));
                    }
                    nPageRef = iPageRef.iterateNext();
                }
                page.pageExtensionList.removeByKey(page.getId()); // remove self
            }

            nPage = iPage.iterateNext();
        }
    }
    else if (/^0\.2\.5\.\d+$/.test(version))
    {
        var iPage = doc.evaluate("./page", nYarip, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
        var nPage = iPage.iterateNext();
        while (nPage)
        {
            pageName = nPage.getAttribute("name");
            var pageId = nPage.getAttribute("id");
            var page = map.obj[pageName] ? map.obj[pageName] : new YaripPage(pageId ? pageId : this.getId(), pageName);

            var iElementChild = doc.evaluate("./element", nPage, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
            var nElementChild = iElementChild.iterateNext();
            if (nElementChild)
            {
                // ELEMENT WHITELIST
                var iElementWhitelist = doc.evaluate("./whitelist", nElementChild, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                var nElementWhitelist = iElementWhitelist.iterateNext();
                if (nElementWhitelist)
                {
                    var iItem = doc.evaluate("./item", nElementWhitelist, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                    var nItem = iItem.iterateNext();
                    while (nItem)
                    {
                        var iXPath = doc.evaluate("./xpath", nItem, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                        var nXPath = iXPath.iterateNext();
                        if (nXPath) {
                            var created = nItem.getAttribute("created");
                            page.elementWhitelist.add(new YaripElementWhitelistItem(
                                nXPath.textContent,
                                nItem.getAttribute("deep"),
                                nItem.getAttribute("force"),
                                created ? created : pageId,
                                nItem.getAttribute("lastFound"),
                                nItem.getAttribute("found"),
                                nItem.getAttribute("notFound"),
                                nItem.getAttribute("ignored")));
                        }
                        nItem = iItem.iterateNext();
                    }
                }

                // ELEMENT BLACKLIST
                var iElementBlacklist = doc.evaluate("./blacklist", nElementChild, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                var nElementBlacklist = iElementBlacklist.iterateNext();
                if (nElementBlacklist)
                {
                    var iItem = doc.evaluate("./item", nElementBlacklist, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                    var nItem = iItem.iterateNext();
                    while (nItem)
                    {
                        var iXPath = doc.evaluate("./xpath", nItem, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                        var nXPath = iXPath.iterateNext();
                        if (nXPath) {
                            var created = nItem.getAttribute("created");
                            var element = new YaripElementBlacklistItem(
                                nXPath.textContent,
                                nItem.getAttribute("deep"),
                                nItem.getAttribute("force"),
                                nItem.getAttribute("dow"),
                                created ? created : pageId,
                                nItem.getAttribute("lastFound"),
                                nItem.getAttribute("found"),
                                nItem.getAttribute("notFound"),
                                nItem.getAttribute("ignored"));
                            page.elementBlacklist.add(element);
                        }
                        nItem = iItem.iterateNext();
                    }
                }

                // ELEMENT ATTRIBUTE
                var iElementAttribute = doc.evaluate("./attribute", nElementChild, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                var nElementAttribute = iElementAttribute.iterateNext();
                if (nElementAttribute)
                {
                    var iItem = doc.evaluate("./item", nElementAttribute, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                    var nItem = iItem.iterateNext();
                    while (nItem)
                    {
                        var iXPath = doc.evaluate("./xpath", nItem, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                        var nXPath = iXPath.iterateNext();
                        var iName = doc.evaluate("./name", nItem, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                        var nName = iName.iterateNext();
                        var iValue = doc.evaluate("./value", nItem, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                        var nValue = iValue.iterateNext();
                        if (nXPath && nName && nValue) {
                            var created = nItem.getAttribute("created");
                            page.elementAttributeList.add(new YaripElementAttributeItem(
                                nXPath.textContent,
                                nName.textContent,
                                nValue.textContent,
                                nItem.getAttribute("deep"),
                                created ? created : pageId,
                                nItem.getAttribute("lastFound"),
                                nItem.getAttribute("found"),
                                nItem.getAttribute("notFound")));
                        }
                        nItem = iItem.iterateNext();
                    }
                }

                // ELEMENT SCRIPT
                var iElementScript = doc.evaluate("./script", nElementChild, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                var nElementScript = iElementScript.iterateNext();
                if (nElementScript)
                {
                    var iItem = doc.evaluate("./item", nElementScript, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                    var nItem = iItem.iterateNext();
                    while (nItem)
                    {
                        var iXPath = doc.evaluate("./xpath", nItem, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                        var nXPath = iXPath.iterateNext();
                        var iScript = doc.evaluate("./script", nItem, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                        var nScript = iScript.iterateNext();
                        if (nXPath) {
                            var created = nItem.getAttribute("created");
                            var element = new YaripElementScriptItem(
                                nXPath.textContent,
                                nScript ? nScript.textContent : null,
                                nItem.getAttribute("deep"),
                                created ? created : pageId,
                                nItem.getAttribute("lastFound"),
                                nItem.getAttribute("found"),
                                nItem.getAttribute("notFound"));
                            page.elementScriptList.add(element);
                        }
                        nItem = iItem.iterateNext();
                    }
                }
            }

            var iContentChild = doc.evaluate("./content", nPage, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
            var nContentChild = iContentChild.iterateNext();
            if (nContentChild)
            {
                // CONTENT WHITELIST
                var iContentWhitelist = doc.evaluate("./whitelist", nContentChild, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                var nContentWhitelist = iContentWhitelist.iterateNext();
                if (nContentWhitelist)
                {
                    page.contentWhitelist.setExclusive(nContentWhitelist.getAttribute("exclusive"));
                    var iItem = doc.evaluate("./item", nContentWhitelist, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                    var nItem = iItem.iterateNext();
                    while (nItem)
                    {
                        var iRegexp = doc.evaluate("./regexp", nItem, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                        var nRegexp = iRegexp.iterateNext();
                        if (nRegexp) {
                            var created = nItem.getAttribute("created");
                            page.contentWhitelist.add(new YaripContentWhitelistItem(
                                nRegexp.textContent,
                                nItem.getAttribute("deep"),
                                nItem.getAttribute("force"),
                                created ? created : pageId,
                                nItem.getAttribute("lastFound")));
                        }
                        nItem = iItem.iterateNext();
                    }
                }

                // CONTENT BLACKLIST
                var iContentBlacklist = doc.evaluate("./blacklist", nContentChild, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                var nContentBlacklist = iContentBlacklist.iterateNext();
                if (nContentBlacklist)
                {
                    var iItem = doc.evaluate("./item", nContentBlacklist, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                    var nItem = iItem.iterateNext();
                    while (nItem)
                    {
                        var iRegexp = doc.evaluate("./regexp", nItem, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                        var nRegexp = iRegexp.iterateNext();
                        if (nRegexp) {
                            var created = nItem.getAttribute("created");
                            page.contentBlacklist.add(new YaripContentBlacklistItem(
                                nRegexp.textContent,
                                nItem.getAttribute("deep"),
                                nItem.getAttribute("force"),
                                created ? created : pageId,
                                nItem.getAttribute("lastFound"),
                                nItem.getAttribute("ignored")));
                        }
                        nItem = iItem.iterateNext();
                    }
                }

                var iContentHeader = doc.evaluate("./header", nContentChild, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                var nContentHeader = iContentHeader.iterateNext();
                if (nContentHeader)
                {
                    // CONTENT REQUEST HEADER
                    var iContentRequestHeader = doc.evaluate("./request", nContentHeader, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                    var nContentRequestHeader = iContentRequestHeader.iterateNext();
                    if (nContentRequestHeader)
                    {
                        var iItem = doc.evaluate("./item", nContentRequestHeader, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                        var nHeader = iItem.iterateNext();
                        while (nHeader)
                        {
                            var iRegexp = doc.evaluate("./regexp", nHeader, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                            var nRegexp = iRegexp.iterateNext();
                            var iName = doc.evaluate("./name", nHeader, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                            var nName = iName.iterateNext();
                            var iScript = doc.evaluate("./script", nHeader, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                            var nScript = iScript.iterateNext();
                            if (nRegexp) {
                                var created = nHeader.getAttribute("created");
                                page.contentRequestHeaderList.add(new YaripHeaderItem(
                                    nRegexp.textContent,
                                    nName.textContent,
                                    nScript ? nScript.textContent : null,
                                    nHeader.getAttribute("merge"),
                                    nHeader.getAttribute("deep"),
                                    created ? created : pageId,
                                    nHeader.getAttribute("lastFound")));
                            }
                            nHeader = iItem.iterateNext();
                        }
                    }

                    // CONTENT RESPONSE HEADER
                    var iContentResponseHeader = doc.evaluate("./response", nContentHeader, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                    var nContentResponseHeader = iContentResponseHeader.iterateNext();
                    if (nContentResponseHeader)
                    {
                        var iItem = doc.evaluate("./item", nContentResponseHeader, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                        var nHeader = iItem.iterateNext();
                        while (nHeader)
                        {
                            var iRegexp = doc.evaluate("./regexp", nHeader, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                            var nRegexp = iRegexp.iterateNext();
                            var iName = doc.evaluate("./name", nHeader, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                            var nName = iName.iterateNext();
                            var iScript = doc.evaluate("./script", nHeader, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                            var nScript = iScript.iterateNext();
                            if (nRegexp) {
                                var created = nHeader.getAttribute("created");
                                page.contentResponseHeaderList.add(new YaripHeaderItem(
                                    nRegexp.textContent,
                                    nName.textContent,
                                    nScript ? nScript.textContent : null,
                                    nHeader.getAttribute("merge"),
                                    nHeader.getAttribute("deep"),
                                    created ? created : pageId,
                                    nHeader.getAttribute("lastFound")));
                            }
                            nHeader = iItem.iterateNext();
                        }
                    }
                }

                // CONTENT REDIRECT
                var iContentRedirect = doc.evaluate("./redirect", nContentChild, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                var nContentRedirect = iContentRedirect.iterateNext();
                if (nContentRedirect)
                {
                    var iItem = doc.evaluate("./item", nContentRedirect, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                    var nItem = iItem.iterateNext();
                    while (nItem)
                    {
                        var iRegexp = doc.evaluate("./regexp", nItem, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                        var nRegexp = iRegexp.iterateNext();
                        var iNewSubStr = doc.evaluate("./newsubstr", nItem, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                        var nNewSubStr = iNewSubStr.iterateNext();
                        if (nRegexp) {
                            var created = nItem.getAttribute("created");
                            page.contentRedirectList.add(new YaripRedirectItem(
                                nRegexp.textContent,
                                nNewSubStr ? nNewSubStr.textContent : null,
                                nItem.getAttribute("deep"),
                                created ? created : pageId,
                                nItem.getAttribute("lastFound")));
                        }
                        nItem = iItem.iterateNext();
                    }
                }
            }

            var iPageChild = doc.evaluate("./page", nPage, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
            var nPageChild = iPageChild.iterateNext();
            if (nPageChild)
            {
                // PAGE STYLE
                var iPageStyle = doc.evaluate("./style", nPageChild, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                var nPageStyle = iPageStyle.iterateNext();
                if (nPageStyle) {
                    page.pageStyleList.setLastChanged(nPageStyle.getAttribute("lastChanged"));
                    page.pageStyleList.set(null, null, nPageStyle.textContent);
                }

                // PAGE SCRIPT
                var iPageScript = doc.evaluate("./script", nPageChild, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                var nPageScript = iPageScript.iterateNext();
                if (nPageScript) {
                    page.pageScriptList.setLastChanged(nPageScript.getAttribute("lastChanged"));
                    page.pageScriptList.set(null, null, nPageScript.textContent);
                }

                var iPageHeader = doc.evaluate("./header", nPageChild, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                var nPageHeader = iPageHeader.iterateNext();
                if (nPageHeader)
                {
                    // PAGE REQUEST HEADER
                    var iPageRequestHeader = doc.evaluate("./request", nPageHeader, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                    var nPageRequestHeader = iPageRequestHeader.iterateNext();
                    if (nPageRequestHeader)
                    {
                        var iItem = doc.evaluate("./item", nPageRequestHeader, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                        var nItem = iItem.iterateNext();
                        while (nItem)
                        {
                            var iRegexp = doc.evaluate("./regexp", nItem, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                            var nRegexp = iRegexp.iterateNext();
                            var iName = doc.evaluate("./name", nItem, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                            var nName = iName.iterateNext();
                            var iScript = doc.evaluate("./script", nItem, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                            var nScript = iScript.iterateNext();
                            if (nRegexp) {
                                var created = nItem.getAttribute("created");
                                page.pageRequestHeaderList.add(new YaripHeaderItem(
                                    nRegexp.textContent,
                                    nName.textContent,
                                    nScript ? nScript.textContent : null,
                                    nItem.getAttribute("merge"),
                                    nItem.getAttribute("deep"),
                                    created ? created : pageId,
                                    nItem.getAttribute("lastFound")));
                            }
                            nItem = iItem.iterateNext();
                        }
                    }

                    // PAGE RESPONSE HEADER
                    var iPageResponseHeader = doc.evaluate("./response", nPageHeader, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                    var nPageResponseHeader = iPageResponseHeader.iterateNext();
                    if (nPageResponseHeader)
                    {
                        var iItem = doc.evaluate("./item", nPageResponseHeader, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                        var nItem = iItem.iterateNext();
                        while (nItem)
                        {
                            var iRegexp = doc.evaluate("./regexp", nItem, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                            var nRegexp = iRegexp.iterateNext();
                            var iName = doc.evaluate("./name", nItem, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                            var nName = iName.iterateNext();
                            var iScript = doc.evaluate("./script", nItem, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                            var nScript = iScript.iterateNext();
                            if (nRegexp) {
                                var created = nItem.getAttribute("created");
                                page.pageResponseHeaderList.add(new YaripHeaderItem(
                                    nRegexp.textContent,
                                    nName.textContent,
                                    nScript ? nScript.textContent : null,
                                    nItem.getAttribute("merge"),
                                    nItem.getAttribute("deep"),
                                    created ? created : pageId,
                                    nItem.getAttribute("lastFound")));
                            }
                            nItem = iItem.iterateNext();
                        }
                    }
                }

                // PAGE REDIRECT
                var iPageRedirect = doc.evaluate("./redirect", nPageChild, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                var nPageRedirect = iPageRedirect.iterateNext();
                if (nPageRedirect)
                {
                    var iItem = doc.evaluate("./item", nPageRedirect, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                    var nItem = iItem.iterateNext();
                    while (nItem)
                    {
                        var iRegexp = doc.evaluate("./regexp", nItem, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                        var nRegexp = iRegexp.iterateNext();
                        var iNewSubStr = doc.evaluate("./newsubstr", nItem, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                        var nNewSubStr = iNewSubStr.iterateNext();
                        if (nRegexp) {
                            var created = nItem.getAttribute("created");
                            page.pageRedirectList.add(new YaripRedirectItem(
                                nRegexp.textContent,
                                nNewSubStr ? nNewSubStr.textContent : null,
                                nItem.getAttribute("deep"),
                                created ? created : pageId,
                                nItem.getAttribute("lastFound")));
                        }
                        nItem = iItem.iterateNext();
                    }
                }
            }

            map.add(page);
            nPage = iPage.iterateNext();
        }

        if (imported) {
            this.map.merge(map);
        } else {
            this.map = map;
        }

        var iPage = doc.evaluate("./page", nYarip, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
        var nPage = iPage.iterateNext();
        while (nPage)
        {
            pageName = nPage.getAttribute("name");
            var pageId = nPage.getAttribute("id");
//            var page = map.obj[pageName];
            var page = this.map.obj[pageName];

            var iPageChild = doc.evaluate("./page", nPage, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
            var nPageChild = iPageChild.iterateNext();
            if (nPageChild)
            {
                // PAGE EXTENSION
                var iExtension = doc.evaluate("./extension", nPageChild, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                var nExtension = iExtension.iterateNext();
                if (nExtension)
                {
                    var iItem = doc.evaluate("./item", nExtension, null, Ci.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE, null);
                    var nItem = iItem.iterateNext();
                    while (nItem)
                    {
                        var pageExtId = nItem.textContent;
                        if (pageExtId && pageExtId != page.getId()) {
                            var created = nItem.getAttribute("created");
                            page.pageExtensionList.add(new YaripPageExtensionItem(
                                pageExtId,
                                nItem.getAttribute("doElements"),
                                nItem.getAttribute("doContents"),
                                nItem.getAttribute("doScripts"),
                                nItem.getAttribute("doHeaders"),
                                nItem.getAttribute("doRedirects"),
                                nItem.getAttribute("doLinks"),
                                created ? created : pageId));
                        }
                        nItem = iItem.iterateNext();
                    }
                    page.pageExtensionList.removeByKey(page.getId()); // remove self
                }
            }

            nPage = iPage.iterateNext();
        }
    }

    if (imported) {
        this.map.merge(map);
    } else {
        this.map = map;
    }

    return pageName;
}

yarip.getDoc = function(file)
{
    if (!file) return null;

    var data = this.getData(file);
    if (!data) return null;

    var dp = Cc['@mozilla.org/xmlextras/domparser;1'].createInstance(Ci.nsIDOMParser);
    return dp.parseFromString(data, "text/xml");
}

yarip.getData = function(file)
{
    if (!file) return "";

    var fis = null;
    var sis = null;
    var data = null;
    try {
        fis = Cc['@mozilla.org/network/file-input-stream;1'].createInstance(Ci.nsIFileInputStream);
        fis.init(file, 0x01, 00004, null);
        sis = Cc['@mozilla.org/scriptableinputstream;1'].createInstance(Ci.nsIScriptableInputStream);
        sis.init(fis);
        data = sis.read(sis.available());
        var suc = Cc["@mozilla.org/intl/scriptableunicodeconverter"].createInstance(Ci.nsIScriptableUnicodeConverter);
        suc.charset = this.CHARSET;
        var data = suc.ConvertToUnicode(data);
    } catch (e) {
    } finally {
        if (sis) try { sis.close(); } catch (e) {}
        if (fis) try { fis.close(); } catch (e) {}
    }
    return data ? data : "";
}

yarip.getFile = function(path)
{
    if (!path) return null;
    var file = Cc["@mozilla.org/file/directory_service;1"].getService(Ci.nsIProperties).get("ProfD", Ci.nsIFile);
    var aPath = path.split(/\/|\\/);
    for (var i = 0; i < aPath.length; i++) file.append(aPath[i]);
    if (!file.exists()) file.create(Ci.nsIFile.NORMAL_FILE_TYPE, 0600);
    if (file.isFile()) return file;
    else return null;
}

yarip.saveToFile = function(data, file)
{
    if (data != "" && !data || !file) return;

    var fos = null;
    var os = null;
    try {
        fos = Cc["@mozilla.org/network/file-output-stream;1"].createInstance(Ci.nsIFileOutputStream);
        fos.init(file, 0x02 | 0x08 | 0x20, 0600, 0);
//        fos.write(data, data.length);
        os = Cc["@mozilla.org/intl/converter-output-stream;1"].createInstance(Ci.nsIConverterOutputStream);
        os.init(fos, this.CHARSET, 0, 0x0000);
        os.writeString(data);
    } catch (e) {
    } finally {
        if (fos) try { fos.close(); } catch (e) {}
        if (os) try { os.close(); } catch (e) {}
    }
}

yarip.logContentLocation = function(status, location, contentLocation, mimeTypeGuess)
{
    var date = new Date();
    for each (var monitor in this.monitorDialogues) {
        monitor.logContentLocation(status, location, contentLocation, date, mimeTypeGuess);
    }
}

yarip.updateContentType = function(status, location, contentLocation, contentType, responseStatus)
{
    for each (var monitor in this.monitorDialogues) {
        monitor.updateContentType(status, location, contentLocation, contentType, responseStatus);
    }
}

//yarip.updateContentLocation = function(status, location, oldContentLocation, newContentLocation)
//{
//    for each (var monitor in this.monitorDialogues) {
////        monitor.logContentLocation(status, location, contentLocation, date, mimeTypeGuess);
//        monitor.updateContentLocation(status, location, oldContentLocation, newContentLocation);
//    }
//}

/*
 * nsIContentPolicy
 */

var TYPE_DOCUMENT = Ci.nsIContentPolicy.TYPE_DOCUMENT;
var ACCEPT = Ci.nsIContentPolicy.ACCEPT;
var REJECT_SERVER = Ci.nsIContentPolicy.REJECT_SERVER;
var DOCUMENT_NODE = Ci.nsIDOMNode.DOCUMENT_NODE;

yarip.shouldLoad = function(contentType, contentLocation, requestOrigin, context, mimeTypeGuess, extra)
{
    if (!this.enabled || !this.schemesRegExp) return ACCEPT;
    if (!context || !contentLocation || !requestOrigin) return ACCEPT;
    if (!this.schemesRegExp.test(contentLocation.scheme)) return ACCEPT;
    if (!this.schemesRegExp.test(requestOrigin.scheme)) return ACCEPT;

    // Getting the document.location.
    if (context.nodeType != DOCUMENT_NODE) context = context.ownerDocument;
    var defaultView = context && context.nodeType == DOCUMENT_NODE ? context.defaultView : null;
    var doc = defaultView ? defaultView.document : null;
    var location = doc ? doc.location : null;
//    if (!location && requestOrigin) location = this.getLocationFromContentLocation(requestOrigin);
    if (!location || !this.schemesRegExp.test(location.protocol.replace(/:$/, ""))) return ACCEPT;

    // Getting the pageName.
    location = this.getLocationFromLocation(location);
    var pageName = this.getPageName(location, this.MODE_PAGE);
    if (!pageName) {
        this.logContentLocation(this.STATUS_UNKNOWN, location, contentLocation, mimeTypeGuess);
        return ACCEPT;
    }

    // Getting the addressObj.
    var addressObj = this.getAddressObj(pageName, true, true); // reuse known addressObj
    if (!addressObj.found) {
        if (this.privateBrowsing) {
            var page = this.createPage(location, null, true);
            this.reloadPage(page.getName());
            addressObj = this.getAddressObj(pageName, true);
        } else {
            this.logContentLocation(this.STATUS_UNKNOWN, location, contentLocation, mimeTypeGuess);
            return ACCEPT;
        }
    }

    var status = this.shouldBlacklist(pageName, addressObj, contentLocation.asciiSpec, defaultView);
    this.logContentLocation(status, location, contentLocation, mimeTypeGuess);
    switch (status) {
    case this.STATUS_UNKNOWN:
        return ACCEPT;
    case this.STATUS_WHITELISTED:
        return ACCEPT;
    case this.STATUS_BLACKLISTED:
        return REJECT_SERVER;
    }
}

yarip.shouldProcess = function(contentType, contentLocation, requestOrigin, context, mimeType, extra)
{
    return ACCEPT;
}

yarip.shouldBlacklist = function(pageName, addressObj, url, defaultView, doFlag)
{
    doFlag = doFlag ? doFlag : 2; // 2 = content
    var whitelisted = false;
    var result = this.STATUS_UNKNOWN;

    // WHITELIST
    for (var a in addressObj.obj)
    {
        if (whitelisted) break;
        if (!addressObj.obj[a][doFlag]) continue; // not doFlag

        var contentWhitelist = this.map.get(a).contentWhitelist;
        if (contentWhitelist.length <= 0) continue;

        var isSelf = addressObj.obj[a][0];
        var list = contentWhitelist.obj;
        for each (var content in list)
        {
            if (isSelf && !content.getDeep() && !this.equalAddresses(pageName, a)) continue;
            if (!content.getRegExpObj().test(url)) continue;

            if (isSelf) {
                content.incrementLastFound();
            }
            if (defaultView) {
                defaultView.yaripStatus = "found";
            }
            if (content.getForce()) {
                return this.STATUS_WHITELISTED;
            } else {
                whitelisted = true;
                result = this.STATUS_WHITELISTED;
                break;
            }
        }
    }

    // BLACKLIST
    for (var a in addressObj.obj)
    {
        if (!addressObj.obj[a][doFlag]) continue; // not doFlag

        var contentBlacklist = this.map.get(a).contentBlacklist;
        if (contentBlacklist.length <= 0) continue;

        var isSelf = addressObj.obj[a][0];
        var list = contentBlacklist.obj;
        for each (var content in list)
        {
            if (isSelf && !content.getDeep() && !this.equalAddresses(pageName, a)) continue;
            if (!content.getRegExpObj().test(url)) continue;
            if (whitelisted && !content.getForce()) {
                if (isSelf) {
                    content.incrementIgnored();
                }
                continue;
            }

            if (isSelf) {
                content.incrementLastFound();
            }
            if (defaultView) {
                defaultView.yaripStatus = "found";
            }
            return this.STATUS_BLACKLISTED;
        }
    }

    return !whitelisted && addressObj.exclusive ? this.STATUS_BLACKLISTED : result;
}

yarip.equalAddresses = function(pageName, address)
{
    if (pageName.indexOf(address) == -1) return true; // ignore deep (extension)

    pageName = pageName.replace(/\/+$/, "");
    address = address.replace(/\/+$/, "");
    if (this.noSchemeRegExp.test(address)) {
        pageName = pageName.replace(/^[^:/?#]+:\/{2}/, ""); // removing scheme
    }

    return pageName == address;
}

yarip.openLocation = function(url)
{
    try {
        var wm = Cc["@mozilla.org/appshell/window-mediator;1"].getService(Ci.nsIWindowMediator);
        var gBrowser = wm.getMostRecentWindow("navigator:browser").gBrowser;
        gBrowser.selectedTab = gBrowser.addTab(url);
    } catch (e) {
        var uri = IOS.newURI(url, this.CHARSET, null);
        var eps = Cc["@mozilla.org/uriloader/external-protocol-service;1"].getService(Ci.nsIExternalProtocolService);
        eps.loadURI(uri);
    }
}

/*
 * nsIObserverService
 */

yarip.observe = function(subject, topic, data)
{
    switch (topic) {
    case "app-startup":
        OS.addObserver(yarip, "profile-after-change", false);
        break;
    case "quit-application":
        yarip.save();
        break;
    case "profile-after-change":
        yarip.map = new YaripMap();
        var file = yarip.getFile(yarip.FILE);
        yarip.load(file);
        yarip.check(file);

        yarip.toggleEnabled(yarip.getValue(yarip.PREF_ENABLED, true, yarip.DATA_TYPE_BOOLEAN));
        yarip.setMode(yarip.getValue(yarip.PREF_MODE, yarip.MODE_FQDN, yarip.DATA_TYPE_INTEGER));
        yarip.setAlwaysUseIndex(yarip.getValue(yarip.PREF_INDEX, false, yarip.DATA_TYPE_BOOLEAN));
        yarip.setElementsInContext(yarip.getValue(yarip.PREF_ELEMENTS, 4, yarip.DATA_TYPE_INTEGER));
        yarip.setPurgeInnerHTML(yarip.getValue(yarip.PREF_PURGE, false, yarip.DATA_TYPE_BOOLEAN));
        yarip.setExclusiveOnCreation(yarip.getValue(yarip.PREF_EXCLUSIVE, false, yarip.DATA_TYPE_BOOLEAN));
        yarip.setTemplates(yarip.getValue(yarip.PREF_TEMPLATES, "", yarip.DATA_TYPE_STRING));
        yarip.setSchemes(yarip.getValue(yarip.PREF_SCHEMES, "^https?$", yarip.DATA_TYPE_STRING));
        yarip.setPrivateBrowsing(yarip.getValue(yarip.PREF_PRIVATE, "false", yarip.DATA_TYPE_BOOLEAN));

        OS.addObserver(yarip, "quit-application", false);
        OS.addObserver(yarip, "http-on-modify-request", false);
        OS.addObserver(yarip, "http-on-examine-response", false);
        OS.addObserver(yarip, "http-on-examine-cached-response", false);
        OS.addObserver(yarip, "http-on-examine-merged-response", false);

        WP.addProgressListener(yarip, Ci.nsIWebProgress.NOTIFY_STATE_DOCUMENT | Ci.nsIWebProgress.NOTIFY_STATE_REQUEST);
        break;
    case "http-on-modify-request":
        if (!yarip.enabled) return;
        if (!(subject instanceof Ci.nsIHttpChannel)) return;

        subject.QueryInterface(Ci.nsIHttpChannel);
        yarip.modifyRequest(subject);
        break;
    case "http-on-examine-response":
    case "http-on-examine-cached-response":
    case "http-on-examine-merged-response":
        if (!yarip.enabled) return;
        if (!(subject instanceof Ci.nsIHttpChannel)) return;

        subject.QueryInterface(Ci.nsIHttpChannel);
        yarip.examineResponse(subject);
        break;
    }
}

yarip.getInterface = function(channel, iid)
{
    if (!channel) return null;

    if (channel.notificationCallbacks) {
        try {
            return channel.notificationCallbacks.QueryInterface(Ci.nsIInterfaceRequestor).getInterface(iid);
        } catch(e) {}
    }

    if (channel.loadGroup) {
        if (channel.loadGroup.notificationCallbacks) {
            try {
                return channel.loadGroup.notificationCallbacks.QueryInterface(Ci.nsIInterfaceRequestor).getInterface(iid);
            } catch(e) {}
        }
        if (channel.loadGroup.groupObserver) {
            try {
                return channel.loadGroup.groupObserver.QueryInterface(Ci.nsIInterfaceRequestor).getInterface(iid);
            } catch(e) {}
        }
    }

    return null;
}

yarip.getLocation = function(channel, doc)
{
    var obj = {
        location: null,
        isPage: false
    };
    try {
        var newURI = IOS.newURI(channel.loadGroup.defaultLoadRequest.name, channel.URI.originCharset, null);
        obj.location = this.getLocationFromContentLocation(newURI);
        return obj;
    } catch (e) {
        if (Ci.nsIChannel.LOAD_DOCUMENT_URI & channel.loadFlags) {
            obj.location = this.getLocationFromContentLocation(channel.URI);
            obj.isPage = true;
            return obj;
        } else if (doc) {
            obj.location = this.getLocationFromLocation(doc.location);
            return obj;
        } else {
            return null;
        }
    }
}

yarip.modifyRequest = function(channel)
{
    if (!this.schemesRegExp.test(channel.URI.scheme)) return;

    var doc = null;
    try { doc = this.getInterface(channel, Ci.nsIDOMWindow).document; } catch(e) {}
    if (!doc) try { doc = this.getInterface(channel, Ci.nsIWebNavigation).document; } catch(e) {}
    var locObj = this.getLocation(channel, doc);
    if (!locObj) return;

    var location = locObj.location;
    var isPage = locObj.isPage;

    var redObj = this.shouldRedirect(location, channel, doc ? doc.defaultView : null, isPage);
    if (!redObj) return;

    switch (redObj.status) {
    case this.STATUS_UNKNOWN:
        if (!isPage) this.logContentLocation(this.STATUS_UNKNOWN, location, channel.URI);
        break;
    case this.STATUS_WHITELISTED:
        if (!isPage) this.logContentLocation(this.STATUS_WHITELISTED, location, channel.URI);
        break;
    case this.STATUS_BLACKLISTED:
        channel.cancel(Cr.NS_ERROR_ABORT);
        /*if (!isPage)*/ this.logContentLocation(this.STATUS_BLACKLISTED, location, channel.URI);
        break;
    case this.STATUS_REDIRECTED:
        new YaripChannelReplacement(channel, redObj.newURI, function(oldChannel, newChannel) {
//                this.logContentLocation(yarip.STATUS_REDIRECTED, location, newChannel.URI);
////                this.updateContentLocation(yarip.STATUS_REDIRECTED, location, oldChannel.URI, newChannel.URI);
            });
        break;
    }
}

yarip.shouldRedirect = function(location, channel, defaultView, isPage)
{
    if (!this.schemesRegExp.test(location.protocol.replace(/:$/, ""))) return null;

    var statusObj = {
        status: this.STATUS_UNKNOWN,
        newURI: null
    };

    var pageName = this.getPageName(location, this.MODE_PAGE);
    if (!pageName) return statusObj;

    var addressObj = this.getAddressObj(pageName, true);
    if (!addressObj.found) return statusObj;

    for (var a in addressObj.obj)
    {
        var isSelf = addressObj.obj[a][0];

        /*
         * REQUEST HEADER
         */

        var headerList = isPage ? this.map.get(a).pageRequestHeaderList : this.map.get(a).contentRequestHeaderList;
        if (addressObj.obj[a][4] && headerList.length > 0)
        {
            for each (var header in headerList.obj)
            {
                if (isSelf && !header.getDeep() && !this.equalAddresses(pageName, a)) continue;
                if (!header.getRegExpObj().test(channel.URI.asciiSpec)) continue;

                var sandbox = new Cu.Sandbox(defaultView ? defaultView : channel.URI.asciiSpec);
                var headerValue = null; // object
                try { headerValue = channel.getRequestHeader(header.getHeaderName()) } catch (e) {}
                try {
                    if (typeof headerValue == "string") {
                        if (/^\s*function\b/.test(header.getScript())) {
                            sandbox.headerValue = headerValue;
                            headerValue = Cu.evalInSandbox("(" + header.getScript() + ")(headerValue);", sandbox);
                        } else {
                            headerValue = header.getScript();
                        }
                    } else {
                        if (/^\s*function\b/.test(header.getScript())) {
                            headerValue = Cu.evalInSandbox("(" + header.getScript() + ")();", sandbox);
                        } else {
                            headerValue = header.getScript();
                        }
                    }

                    if (typeof headerValue != "string") {
                        Cu.reportError("Yarip: Function doesn't return a string!\nPage: " + a + " > "+ (isPage ? "Page" : "Content") + " > Header > Request: " + header.getHeaderName());
                        continue;
                    }

                    channel.setRequestHeader(header.getHeaderName(), headerValue, header.getMerge());

                    if (isSelf) {
                        header.incrementLastFound();
                    }

                    if (defaultView) defaultView.yaripStatus = "found";
                } catch (e) {
                    Cu.reportError("Yarip: Couldn't set request-header!\nPage: " + a + " > "+ (isPage ? "Page" : "Content") + " > Header > Request: " + header.getHeaderName() + "\n*** e=" + e);
//                    CS.logMessage(message);
//                    CS.logStringMessage(message);
                }
            }
        }

        /*
         * REDIRECT
         */

        var redirectList = isPage ? this.map.get(a).pageRedirectList : this.map.get(a).contentRedirectList;
        if (addressObj.obj[a][5] && redirectList.length > 0)
        {
            for each (var redirect in redirectList.obj)
            {
                if (isSelf && !redirect.getDeep() && !this.equalAddresses(pageName, a)) continue;
                if (!redirect.getRegExpObj().test(channel.URI.asciiSpec)) continue;

//                try
//                {
                    var newSpec = channel.URI.asciiSpec.replace(redirect.getRegExpObj(), redirect.getNewSubStr());
                    var newURI = IOS.newURI(newSpec, channel.URI.originCharset, null);

                    if (isSelf) {
                        redirect.incrementLastFound();
                    }

                    if (defaultView) defaultView.yaripStatus = "found";

                    statusObj.status = this.STATUS_REDIRECTED;
                    statusObj.newURI = newURI;
                    return statusObj;
//                } catch (e) {
//                    Cu.reportError("Yarip: Couldn't redirect!\n" + e + "\n" + a + ": " + newSpec);
////                    CS.logMessage(message);
////                    CS.logStringMessage(message);
//                }
            }
        }
    }

    if (isPage) {
        statusObj.status = this.STATUS_WHITELISTED;
        return statusObj;
    } else {
        statusObj.status = this.shouldBlacklist(pageName, addressObj, channel.URI.asciiSpec, defaultView);
        return statusObj;
    }
}

yarip.examineResponse = function(channel)
{
    if (!this.schemesRegExp.test(channel.URI.scheme)) return;

    var win = null;
    var doc = null;
    var defaultView = null;
    try {
        win = this.getInterface(channel, Ci.nsIDOMWindow);
        doc = win.document;
        defaultView = doc.defaultView;
    } catch(e) {}
    var locObj = this.getLocation(channel, doc);
    if (!locObj) return;

    var location = locObj.location;
    var isPage = locObj.isPage;
    if (!this.schemesRegExp.test(location.protocol.replace(/:$/, ""))) return;

    this.updateContentType(null, location, channel.URI, channel.contentType, channel.responseStatus);

    var pageName = this.getPageName(location, this.MODE_PAGE);
    if (!pageName) return;

    var addressObj = this.getAddressObj(pageName, true);
    if (!addressObj.found) return;

    /*
     * BLACKLIST
     */

    var status = channel.responseStatus;
    if (status >= 300 && status < 400)
    {
        var isLinking = isPage || (Ci.nsIChannel.LOAD_INITIAL_DOCUMENT_URI & channel.loadFlags);
        if (!isLinking || this.privateBrowsing)
        {
            var newURI = null;
            try { newURI = IOS.newURI(channel.getResponseHeader("Location"), channel.URI.originCharset, null); } catch (e) {}
            if (newURI) {
                var status = this.shouldBlacklist(pageName, addressObj, newURI.asciiSpec, defaultView, isLinking ? 6 : 2 /* link or content */);
                switch (status) {
                case this.STATUS_BLACKLISTED:
                    channel.cancel(Cr.NS_ERROR_ABORT);
                    /*if (!isPage)*/ this.logContentLocation(this.STATUS_BLACKLISTED, location, newURI);
                    return;
                }
            }
        }
    }

    for (var a in addressObj.obj)
    {
        if (!addressObj.obj[a][4]) continue; // not header

        var isSelf = addressObj.obj[a][0];

        /*
         * RESPONSE HEADER
         */
        var headerList = isPage ? this.map.get(a).pageResponseHeaderList : this.map.get(a).contentResponseHeaderList;
        if (headerList.length > 0)
        {
            for each (var header in headerList.obj)
            {
                if (isSelf && !header.getDeep() && !this.equalAddresses(pageName, a)) continue;
                if (!header.getRegExpObj().test(channel.URI.asciiSpec)) continue;

                var sandbox = new Cu.Sandbox(win ? win : location.asciiHref);
                var headerValue = null; // object
                try { headerValue = channel.getResponseHeader(header.getHeaderName()); } catch (e) {}
                try {
                    if (typeof headerValue == "string") {
                        if (/^\s*function\b/.test(header.getScript())) {
                            sandbox.headerValue = headerValue;
                            headerValue = Cu.evalInSandbox("(" + header.getScript() + ")(headerValue);", sandbox);
                        } else {
                            headerValue = header.getScript();
                        }
                    } else {
                        if (/^\s*function\b/.test(header.getScript())) {
                            headerValue = Cu.evalInSandbox("(" + header.getScript() + ")();", sandbox);
                        } else {
                            headerValue = header.getScript();
                        }
                    }

                    if (typeof headerValue != "string") {
                        Cu.reportError("Yarip: Function doesn't return a string!\nPage: " + a + " > "+ (isPage ? "Page" : "Content") + " > Header > Response: " + header.getHeaderName());
                        continue;
                    }

                    channel.setResponseHeader(header.getHeaderName(), headerValue, header.getMerge());

                    if (isSelf) {
                        header.incrementLastFound();
                    }
                } catch (e) {
                    Cu.reportError("Yarip: Couldn't set request-header!\nPage: " + a + " > "+ (isPage ? "Page" : "Content") + " > Header > Response: " + header.getHeaderName() + "\n*** e=" + e);
//                    CS.logMessage(message);
//                    CS.logStringMessage(message);
                }
            }
        }
    }
}

/*
 * nsIWebProgressListener
 */

var WPL = Ci.nsIWebProgressListener;

yarip.onStateChange = function(webProgress, request, stateFlags, _status)
{
    if (!this.enabled) return;
    if (!webProgress.isLoadingDocument) return;
    if (!(request instanceof Ci.nsIHttpChannel)) return;
    if (WPL.STATE_STOP & stateFlags) return;

    request.QueryInterface(Ci.nsIHttpChannel);
    if (!this.schemesRegExp.test(request.URI.scheme)) return;

    var doc = webProgress.DOMWindow.document;
    if (!this.schemesRegExp.test(doc.location.protocol.replace(/:$/, ""))) return;

    var isLinking = (Ci.nsIChannel.LOAD_INITIAL_DOCUMENT_URI & request.loadFlags) || (WPL.STATE_REDIRECTING & stateFlags);
    if (isLinking && !this.privateBrowsing) return;

    var location = this.getLocationFromLocation(doc.location);
    var pageName = this.getPageName(location, this.MODE_PAGE);
    if (!pageName) {
        this.logContentLocation(this.STATUS_UNKNOWN, location, request.URI);
        return;
    }

    var addressObj = this.getAddressObj(pageName, true);
    if (!addressObj.found) {
        this.logContentLocation(this.STATUS_UNKNOWN, location, request.URI);
        return;
    }

    var status = this.shouldBlacklist(pageName, addressObj, request.URI.asciiSpec, doc.defaultView, isLinking ? 6 : 2 /* link or content */);
    var isPage = Ci.nsIChannel.LOAD_DOCUMENT_URI & request.loadFlags;
    switch (status) {
    case this.STATUS_UNKNOWN:
        if (!isPage) this.logContentLocation(this.STATUS_UNKNOWN, location, request.URI);
        break;
    case this.STATUS_WHITELISTED:
        if (!isPage) this.logContentLocation(this.STATUS_WHITELISTED, location, request.URI);
        break;
    case this.STATUS_BLACKLISTED:
        request.cancel(Cr.NS_ERROR_ABORT);
        /*if (!isPage)*/ this.logContentLocation(this.STATUS_BLACKLISTED, location, request.URI);
        break;
    }
}

/*
 * nsISupportsWeakReference
 */

yarip.GetWeakReference = function()
{
    return Cu.getWeakReference(this);
}

/*
 * nsIYarip
 */

yarip.classID = Components.ID("{46d09504-6a0b-40c8-8ce8-d07cf370d25e}");
yarip.classDescription = "Yet Another Remove It Permanently";
yarip.contractID = "@yarip.mozdev.org;1";
yarip._xpcom_categories = [
    { category: "app-startup", service: true },
    { category: "content-policy" }
];
yarip.QueryInterface = XPCOMUtils.generateQI([
    Ci.nsISupports,
    Ci.nsIContentPolicy,
    Ci.nsIObserver,
    Ci.nsIWebProgressListener,
    Ci.nsISupportsWeakReference
]);

yarip.wrappedJSObject = yarip;
function yaripComponent() {}
yaripComponent.prototype = yarip;

/**
 * XPCOMUtils.generateNSGetFactory was introduced in Mozilla 2 (Firefox 4).
 * XPCOMUtils.generateNSGetModule is for Mozilla 1.9.2 (Firefox 3.6).
 */

if (XPCOMUtils.generateNSGetFactory) {
    var NSGetFactory = XPCOMUtils.generateNSGetFactory([yaripComponent]);
} else {
    var NSGetModule = XPCOMUtils.generateNSGetModule([yaripComponent]);
}
