
/*
Copyright 2007-2011 Kim A. Brandt <kimabrandt@gmx.de>

This file is part of yarip.

Yarip is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

Yarip is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with yarip; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

function YaripItem(xpath, script, deep, force, dow, created, lastFound, found, notFound, ignored)
{
    this.xpath = "";
    this.script = "";
    this.deep = true;
    this.force = true;
    this.dow = false;
    this.created = -1;
    this.lastFound = -1;
    this.found = 0;
    this.notFound = 0;
    this.ignored = 0;

    this.getKey = function()
    {
        return this.getXPath();
    }

    this.setXPath = function(value)
    {
        if (!value) return;
        this.xpath = "" + value;
        this.setCreated(Date.now());
    }

    this.getXPath = function()
    {
        return this.xpath;
    }

    this.setScript = function(value)
    {
        if (!value) return;
        this.script = value;
    }

    this.getScript = function()
    {
        return this.script;
    }

    this.setDeep = function(value)
    {
        if (!value) return;
        this.deep = "" + value != "false";
    }

    this.getDeep = function()
    {
        return this.deep;
    }

    this.setForce = function(value)
    {
        this.force = "" + value != "false";
    }

    this.getForce = function()
    {
        return this.force;
    }

    this.setDOW = function(value)
    {
        this.dow = "" + value == "true";
    }

    this.getDOW = function()
    {
        return this.dow;
    }

    this.setCreated = function(value)
    {
        if (!value) return;
        value = Number(value);
        if (typeof value != "number" || value + "" == "NaN") return;
        this.created = value;
    }

    this.getCreated = function()
    {
        return this.created;
    }

    this.setLastFound = function(value)
    {
        if (!value) return;
        value = Number(value);
        if (typeof value != "number" || value + "" == "NaN") return;
        this.lastFound = value;
    }

    this.getLastFound = function()
    {
        return this.lastFound;
    }

    this.setFound = function(value)
    {
        if (!value) return;
        value = Number(value);
        if (typeof value != "number" || value + "" == "NaN") return;
        this.found = value;
    }

    this.getFound = function()
    {
        return this.found;
    }

    this.incrementFound = function()
    {
        this.found++;
        this.lastFound = Date.now();
    }

    this.setNotFound = function(value)
    {
        if (!value) return;
        value = Number(value);
        if (typeof value != "number" || value + "" == "NaN") return;
        this.notFound = value;
    }

    this.getNotFound = function()
    {
        return this.notFound;
    }

    this.incrementNotFound = function()
    {
        this.notFound++;
    }

    this.setIgnored = function(value)
    {
        if (!value) return;
        value = Number(value);
        if (typeof value != "number" || value + "" == "NaN") return;
        this.ignored = value;
    }

    this.getIgnored = function()
    {
        return this.ignored;
    }

    this.incrementIgnored = function()
    {
        this.ignored++;
    }

    this.purge = function()
    {
        this.created = Date.now();
        this.lastFound = -1;
        this.found = 0;
        this.notFound = 0;
        this.ignored = 0;
    }

    this.clone = function(purge)
    {
        var tmp = new this.constructor(this.xpath, this.script, this.deep, this.force, this.dow, this.created, this.lastFound, this.found, this.notFound, this.ignored);
//        for (var k in this) if (this[k] && typeof this[k] != "object" && !/^(xpath|deep|force|dow|created|lastFound|found|notFound|ignored)$/.test(k)) tmp[k] = this[k];
        if (purge) tmp.purge();
        return tmp;
    }

    this.merge = function(o)
    {
        if (!o) return;
        this.setDeep(o.getDeep() && this.getDeep());
        this.setForce(o.getForce() && this.getForce());
        this.setDOW(o.getDOW() || this.getDOW());
        if (this.getCreated() == -1 || o.getCreated() < this.getCreated()) this.setCreated(o.getCreated());
    }

    this.compare = function(b)
    {
        var aKey = this.getKey();
        var bKey = b.getKey();
        if (aKey < bKey) return -1;
        if (aKey > bKey) return 1;
        return 0;
    }

    this.generateXml = function()
    {
        if (!this.xpath) return "";
        return "\t\t\t\t" +
            "<item" +
                " deep=\"" + this.deep + "\"" +
                " force=\"" + this.force + "\"" +
                " dow=\"" + this.dow + "\"" +
                (this.created > -1 ? " created=\"" + this.created + "\"" : "") +
                (this.lastFound > -1 ? " lastFound=\"" + this.lastFound + "\"" : "") +
                (this.found > 0 ? " found=\"" + this.found + "\"" : "") +
                (this.notFound > 0 ? " notFound=\"" + this.notFound + "\"" : "") +
                (this.ignored > 0 ? " ignored=\"" + this.ignored + "\"" : "") +
            ">\n" +
            "\t\t\t\t\t<xpath><![CDATA[" + this.xpath + "]]></xpath>\n" +
            "\t\t\t\t</item>\n";
    }

    this.setXPath(xpath);
    this.setScript(script);
    this.setDeep(deep);
    this.setForce(force);
    this.setDOW(dow);
    this.setCreated(created ? created : Date.now());
    this.setLastFound(lastFound);
    this.setFound(found);
    this.setNotFound(notFound);
    this.setIgnored(ignored);
}

//function YaripElementWhitelistItem(xpath, deep, created, lastFound, found, notFound, ignored)
function YaripElementWhitelistItem(xpath, deep, force, created, lastFound, found, notFound, ignored)
{
    this.xpath = "";
    this.deep = true;
    this.force = false;
    this.created = -1;
    this.lastFound = -1;
    this.found = 0;
    this.notFound = 0;
    this.ignored = 0;

    this.setXPath(xpath);
    this.setDeep(deep);
    this.setForce(force);
    this.setCreated(created ? created : Date.now());
    this.setLastFound(lastFound);
    this.setFound(found);
    this.setNotFound(notFound);
    this.setIgnored(ignored);
}
YaripElementWhitelistItem.prototype = new YaripItem;
YaripElementWhitelistItem.prototype.constructor = YaripElementWhitelistItem;
YaripElementWhitelistItem.prototype.setForce = function(value)
{
    this.force = "" + value == "true";
}
YaripElementWhitelistItem.prototype.clone = function(purge)
{
    var tmp = new this.constructor(this.xpath, this.deep, this.force, this.created, this.lastFound, this.found, this.notFound, this.ignored);
//    for (var k in this) if (this[k] && typeof this[k] != "object" && !/^(xpath|deep|force|created|lastFound|found|notFound|ignored)$/.test(k)) tmp[k] = this[k];
    if (purge) tmp.purge();
    return tmp;
}
YaripElementWhitelistItem.prototype.merge = function(o)
{
    if (!o) return;
    this.setDeep(o.getDeep() && this.getDeep());
    this.setForce(o.getForce() || this.getForce());
    if (this.getCreated() == -1 || o.getCreated() < this.getCreated()) this.setCreated(o.getCreated());
}
YaripElementWhitelistItem.prototype.generateXml = function()
{
    if (!this.xpath) return "";
    return "\t\t\t\t" +
        "<item" +
            " deep=\"" + this.deep + "\"" +
            " force=\"" + this.force + "\"" +
            (this.created > -1 ? " created=\"" + this.created + "\"" : "") +
            (this.lastFound > -1 ? " lastFound=\"" + this.lastFound + "\"" : "") +
            (this.found > 0 ? " found=\"" + this.found + "\"" : "") +
            (this.notFound > 0 ? " notFound=\"" + this.notFound + "\"" : "") +
            (this.ignored > 0 ? " ignored=\"" + this.ignored + "\"" : "") +
        ">\n" +
        "\t\t\t\t\t<xpath><![CDATA[" + this.xpath + "]]></xpath>\n" +
        "\t\t\t\t</item>\n";
}

function YaripElementBlacklistItem(xpath, deep, force, dow, created, lastFound, found, notFound, ignored)
{
    this.xpath = "";
    this.deep = true;
    this.force = true;
    this.dow = false;
    this.created = -1;
    this.lastFound = -1;
    this.found = 0;
    this.notFound = 0;
    this.ignored = 0;

    this.setDeep(deep);
    this.setForce(force);
    this.setDOW(dow);
    this.setCreated(created ? created : Date.now());
    this.setLastFound(lastFound);
    this.setFound(found);
    this.setNotFound(notFound);
    this.setIgnored(ignored);
    this.setXPath(xpath);
}
YaripElementBlacklistItem.prototype = new YaripItem;
YaripElementBlacklistItem.prototype.setXPath = function(value)
{
    if (!value) return;
    this.xpath = "" + value;
}

function YaripElementAttributeItem(xpath, name, value, deep, created, lastFound, found, notFound)
{
    this.xpath = "";
    this.name = "";
    this.value = "";
    this.deep = true;
    this.created = -1;
    this.lastFound = -1;
    this.found = 0;
    this.notFound = 0;

    this.getKey = function()
    {
        return this.getXPath() + " " + this.getName();
    }

    this.setXPath = function(value)
    {
        if (!value) return;
        this.xpath = "" + value;
        this.setCreated(Date.now());
    }

    this.getXPath = function()
    {
        return this.xpath;
    }

    this.setName = function(value)
    {
        if (!value) return;
        this.name = value;
        this.setCreated(Date.now());
    }

    this.getName = function()
    {
        return this.name;
    }

    this.setValue = function(value)
    {
        if (value) this.value = value;
    }

    this.getValue = function()
    {
        return this.value;
    }

    this.setDeep = function(value)
    {
        if (!value) return;
        this.deep = "" + value != "false";
    }

    this.getDeep = function()
    {
        return this.deep;
    }

    this.setCreated = function(value)
    {
        if (!value) return;
        value = Number(value);
        if (typeof value != "number" || value + "" == "NaN") return;
        this.created = value;
    }

    this.getCreated = function()
    {
        return this.created;
    }

    this.setLastFound = function(value)
    {
        if (!value) return;
        value = Number(value);
        if (typeof value != "number" || value + "" == "NaN") return;
        this.lastFound = value;
    }

    this.getLastFound = function()
    {
        return this.lastFound;
    }

    this.setFound = function(value)
    {
        if (!value) return;
        value = Number(value);
        if (typeof value != "number" || value + "" == "NaN") return;
        this.found = value;
    }

    this.getFound = function()
    {
        return this.found;
    }

    this.incrementFound = function()
    {
        this.found++;
        this.lastFound = Date.now();
    }

    this.setNotFound = function(value)
    {
        if (!value) return;
        value = Number(value);
        if (typeof value != "number" || value + "" == "NaN") return;
        this.notFound = value;
    }

    this.getNotFound = function()
    {
        return this.notFound;
    }

    this.incrementNotFound = function()
    {
        this.notFound++;
    }

    this.purge = function()
    {
        this.created = Date.now();
        this.lastFound = -1;
        this.found = 0;
        this.notFound = 0;
    }

    this.clone = function(purge)
    {
        var tmp = new this.constructor(this.xpath, this.name, this.value, this.deep, this.created, this.lastFound, this.found, this.notFound);
//        for (var k in this) if (this[k] && typeof this[k] != "object" && !/^(xpath|name|value|deep|created|lastFound|found|notFound)$/.test(k)) tmp[k] = this[k];
        if (purge) tmp.purge();
        return tmp;
    }

    this.merge = function(o)
    {
        if (!o) return;
        this.setName(o.getName() || this.getName());
        this.setValue(o.getValue() || this.getValue());
        this.setDeep(o.getDeep() && this.getDeep());
        if (this.getCreated() == -1 || o.getCreated() < this.getCreated()) this.setCreated(o.getCreated());
    }

    this.compare = function(b)
    {
        var aKey = this.getKey();
        var bKey = b.getKey();
        if (aKey < bKey) return -1;
        if (aKey > bKey) return 1;
        return 0;
    }

    this.generateXml = function()
    {
        if (!this.xpath || !this.name || !this.value) return "";
        var tmp = "\t\t\t\t" +
            "<item" +
                " deep=\"" + this.deep + "\"" +
                (this.created > -1 ? " created=\"" + this.created + "\"" : "") +
                (this.lastFound > -1 ? " lastFound=\"" + this.lastFound + "\"" : "") +
                (this.found > 0 ? " found=\"" + this.found + "\"" : "") +
                (this.notFound > 0 ? " notFound=\"" + this.notFound + "\"" : "") +
            ">\n" +
            "\t\t\t\t\t<xpath><![CDATA[" + this.xpath + "]]></xpath>\n" +
            "\t\t\t\t\t<name><![CDATA[" + this.name + "]]></name>\n" +
            "\t\t\t\t\t<value><![CDATA[" + this.value + "]]></value>\n" +
            "\t\t\t\t</item>\n";
        return tmp;
    }

    this.setName(name);
    this.setValue(value);
    this.setDeep(deep);
    this.setCreated(created ? created : Date.now());
    this.setLastFound(lastFound);
    this.setFound(found);
    this.setNotFound(notFound);
    this.setXPath(xpath);
}

function YaripElementScriptItem(xpath, script, deep, created, lastFound, found, notFound)
{
    this.xpath = "";
    this.script = "";
    this.deep = true;
    this.created = -1;
    this.lastFound = -1;
    this.found = 0;
    this.notFound = 0;

    this.setXPath(xpath);
    this.setScript(script);
    this.setDeep(deep);
    this.setCreated(created);
    this.setLastFound(lastFound);
    this.setFound(found);
    this.setNotFound(notFound);
}
YaripElementScriptItem.prototype = new YaripItem;
YaripElementScriptItem.prototype.constructor = YaripElementScriptItem;
YaripElementScriptItem.prototype.clone = function(purge)
{
    var tmp = new this.constructor(this.xpath, this.script, this.deep, this.created, this.lastFound, this.found, this.notFound);
//    for (var k in this) if (this[k] && typeof this[k] != "object" && !/^(xpath|script|deep|created|lastFound|found|notFound)$/.test(k)) tmp[k] = this[k];
    if (purge) tmp.purge();
    return tmp;
}
YaripElementScriptItem.prototype.merge = function(o)
{
    if (!o) return;
    this.setDeep(o.getDeep() && this.getDeep());
    if (this.getCreated() == -1 || o.getCreated() < this.getCreated()) this.setCreated(o.getCreated());
}
YaripElementScriptItem.prototype.generateXml = function()
{
    if (!this.xpath) return "";
    return "\t\t\t\t" +
        "<item" +
            " deep=\"" + this.deep + "\"" +
            (this.created > -1 ? " created=\"" + this.created + "\"" : "") +
            (this.lastFound > -1 ? " lastFound=\"" + this.lastFound + "\"" : "") +
            (this.found > 0 ? " found=\"" + this.found + "\"" : "") +
            (this.notFound > 0 ? " notFound=\"" + this.notFound + "\"" : "") +
        ">\n" +
        "\t\t\t\t\t<xpath><![CDATA[" + this.xpath + "]]></xpath>\n" +
        "\t\t\t\t\t<script><![CDATA[" + this.script + "]]></script>\n" +
        "\t\t\t\t</item>\n";
}

function YaripContentItem(regexp, deep, force, created, lastFound, ignored)
{
    this.regexp = "";
    this.regexpobj = null;
    this.deep = true;
    this.force = true;
    this.created = -1;
    this.lastFound = -1;
    this.ignored = 0;

    this.getKey = function()
    {
        return this.getRegExp();
    }

    this.setRegExp = function(value)
    {
        if (!value) return;

        this.regexp = value;
        this.setRegExpObj(value);
        this.setCreated(Date.now());
    }

    this.getRegExp = function()
    {
        return this.regexp;
    }

    this.setRegExpObj = function(value)
    {
        if (value) this.regexpobj = new RegExp(value);
    }

    this.getRegExpObj = function()
    {
        return this.regexpobj;
    }

    this.setDeep = function(value)
    {
        this.deep = "" + value != "false";
    }

    this.getDeep = function()
    {
        return this.deep;
    }

    this.setForce = function(value)
    {
        this.force = "" + value != "false";
    }

    this.getForce = function()
    {
        return this.force;
    }

    this.setCreated = function(value)
    {
        if (!value) return;
        value = Number(value);
        if (typeof value != "number" || value + "" == "NaN") return;
        this.created = value;
    }

    this.getCreated = function()
    {
        return this.created;
    }

    this.setLastFound = function(value)
    {
        if (!value) return;
        value = Number(value);
        if (typeof value != "number" || value + "" == "NaN") return;
        this.lastFound = value;
    }

    this.getLastFound = function()
    {
        return this.lastFound;
    }

    this.incrementLastFound = function()
    {
        this.lastFound = Date.now();
    }

    this.setIgnored = function(value)
    {
        if (!value) return;
        value = Number(value);
        if (typeof value != "number" || value + "" == "NaN") return;
        this.ignored = value;
    }

    this.getIgnored = function()
    {
        return this.ignored;
    }

    this.incrementIgnored = function()
    {
        this.ignored++;
    }

    this.purge = function()
    {
        this.created = Date.now();
        this.lastFound = -1;
        this.ignored = 0;
    }

    this.clone = function(purge)
    {
        var tmp = new this.constructor(this.regexp, this.deep, this.force, this.created, this.lastFound, this.ignored);
//        for (var k in this) if (this[k] && typeof this[k] != "object" && !/^(regexp|deep|force|created|lastFound|ignored)$/.test(k)) tmp[k] = this[k];
        if (purge) tmp.purge();
        return tmp;
    }

    this.merge = function(o)
    {
        if (!o) return;
        this.setRegExp(o.getRegExp() || this.getRegExp());
        this.setDeep(o.getDeep() && this.getDeep());
        this.setForce(o.getForce() && this.getForce());
        if (this.getCreated() == -1 || o.getCreated() < this.getCreated()) this.setCreated(o.getCreated());
    }

    this.compare = function(b)
    {
        var aKey = this.getKey();
        var bKey = b.getKey();
        if (aKey < bKey) return -1;
        if (aKey > bKey) return 1;
        return 0;
    }

    this.generateXml = function()
    {
        if (!this.regexp) return "";
        return "\t\t\t\t" +
            "<item" +
                " deep=\"" + this.deep + "\"" +
                " force=\"" + this.force + "\"" +
                (this.created > -1 ? " created=\"" + this.created + "\"" : "") +
                (this.lastFound > -1 ? " lastFound=\"" + this.lastFound + "\"" : "") +
                (this.ignored > 0 ? " ignored=\"" + this.ignored + "\"" : "") +
            ">\n" +
            "\t\t\t\t\t<regexp><![CDATA[" + this.regexp + "]]></regexp>\n" +
            "\t\t\t\t</item>\n";
    }

    this.setRegExp(regexp);
    this.setDeep(deep);
    this.setForce(force);
    this.setCreated(created ? created : Date.now());
    this.setLastFound(lastFound);
    this.setIgnored(ignored);
}

//function YaripContentWhitelistItem(regexp, deep, created, lastFound)
function YaripContentWhitelistItem(regexp, deep, force, created, lastFound)
{
    this.regexp = "";
    this.regexpobj = null;
    this.deep = true;
    this.force = false;
    this.created = -1;
    this.lastFound = -1;

    this.setRegExp(regexp);
    this.setDeep(deep);
    this.setForce(force);
    this.setCreated(created ? created : Date.now());
    this.setLastFound(lastFound);
}
YaripContentWhitelistItem.prototype = new YaripContentItem;
YaripContentWhitelistItem.prototype.constructor = YaripContentWhitelistItem;
YaripContentWhitelistItem.prototype.setForce = function(value)
{
    this.force = "" + value == "true";
}
YaripContentWhitelistItem.prototype.clone = function(purge)
{
    var tmp = new this.constructor(this.regexp, this.deep, this.force, this.created, this.lastFound);
//    for (var k in this) if (this[k] && typeof this[k] != "object" && !/^(regexp|deep|force|created|lastFound)$/.test(k)) tmp[k] = this[k];
    if (purge) tmp.purge();
    return tmp;
}
YaripContentWhitelistItem.prototype.merge = function(o)
{
    if (!o) return;
    this.setRegExp(o.getRegExp() || this.getRegExp());
    this.setDeep(o.getDeep() && this.getDeep());
    this.setForce(o.getForce() || this.getForce());
    if (this.getCreated() == -1 || o.getCreated() < this.getCreated()) this.setCreated(o.getCreated());
}
YaripContentWhitelistItem.prototype.generateXml = function()
{
    if (!this.regexp) return "";
    return "\t\t\t\t" +
        "<item" +
            " deep=\"" + this.deep + "\"" +
            " force=\"" + this.force + "\"" +
            (this.created > -1 ? " created=\"" + this.created + "\"" : "") +
            (this.lastFound > -1 ? " lastFound=\"" + this.lastFound + "\"" : "") +
        ">\n" +
        "\t\t\t\t\t<regexp><![CDATA[" + this.regexp + "]]></regexp>\n" +
        "\t\t\t\t</item>\n";
}

function YaripContentBlacklistItem(regexp, deep, force, created, lastFound, ignored)
{
    this.regexp = "";
    this.regexpobj = null;
    this.deep = true;
    this.force = true;
    this.created = -1;
    this.lastFound = -1;
    this.ignored = 0;

    this.setRegExp(regexp);
    this.setDeep(deep);
    this.setForce(force);
    this.setCreated(created ? created : Date.now());
    this.setLastFound(lastFound);
    this.setIgnored(ignored);
}
YaripContentBlacklistItem.prototype = new YaripContentItem;
YaripContentBlacklistItem.prototype.constructor = YaripContentBlacklistItem;

function YaripHeaderItem(regexp, headerName, script, merge, deep, created, lastFound)
{
    this.regexp = "";
    this.regexpobj = null;
    this.headerName = null;
    this.script = "";
    this._merge = false;
    this.deep = true;
    this.created = -1;
    this.lastFound = -1;

    this.getKey = function()
    {
        return this.getRegExp() + " " + this.getHeaderName();
    }

    this.setRegExp = function(value)
    {
        if (!value) return;

        this.regexp = value;
        this.setRegExpObj(value);
        this.setCreated(Date.now());
    }

    this.getRegExp = function()
    {
        return this.regexp;
    }

    this.setRegExpObj = function(value)
    {
        if (value) this.regexpobj = new RegExp(value);
    }

    this.getRegExpObj = function()
    {
        return this.regexpobj;
    }

    this.setHeaderName = function(value)
    {
        if (!value && value != "") return;

        this.headerName = value;
        this.setCreated(Date.now());
    }

    this.getHeaderName = function()
    {
        return this.headerName;
    }

    this.setScript = function(value)
    {
        if (!value && value != "") return;

        this.script = value;
    }

    this.getScript = function()
    {
        return this.script;
    }

    this.setMerge = function(value)
    {
        this._merge = "" + value == "true";
    }

    this.getMerge = function()
    {
        return this._merge;
    }

    this.setDeep = function(value)
    {
        this.deep = "" + value != "false";
    }

    this.getDeep = function()
    {
        return this.deep;
    }

    this.setCreated = function(value)
    {
        if (!value) return;
        value = Number(value);
        if (typeof value != "number" || value + "" == "NaN") return;
        this.created = value;
    }

    this.getCreated = function()
    {
        return this.created;
    }

    this.setLastFound = function(value)
    {
        if (!value) return;
        value = Number(value);
        if (typeof value != "number" || value + "" == "NaN") return;
        this.lastFound = value;
    }

    this.getLastFound = function()
    {
        return this.lastFound;
    }

    this.incrementLastFound = function()
    {
        this.lastFound = Date.now();
    }

    this.purge = function()
    {
        this.created = Date.now();
        this.lastFound = -1;
    }

    this.clone = function(purge)
    {
        var tmp = new this.constructor(this.regexp, this.headerName, this.script, this._merge, this.deep, this.created, this.lastFound);
//        for (var k in this) if (this[k] && typeof this[k] != "object" && !/^(regexp|headerName|script|_merge|deep|created|lastFound)$/.test(k)) tmp[k] = this[k];
        if (purge) tmp.purge();
        return tmp;
    }

    this.merge = function(o)
    {
        if (!o) return;
        this.setRegExp(o.getRegExp() || this.getRegExp());
        this.setHeaderName(o.getHeaderName() || this.getHeaderName());
        this.setMerge(o.getMerge() || this.getMerge());
        this.setDeep(o.getDeep() && this.getDeep());
        if (this.getCreated() == -1 || o.getCreated() < this.getCreated()) this.setCreated(o.getCreated());
    }

    this.compare = function(b)
    {
        var aKey = this.getKey();
        var bKey = b.getKey();
        if (aKey < bKey) return -1;
        if (aKey > bKey) return 1;
        return 0;
    }

    this.generateXml = function()
    {
        if (!this.regexp || !this.getHeaderName()) return "";
        return "\t\t\t\t\t" +
            "<item" +
                (this._merge ? " merge=\"" + this._merge + "\"" : "") +
                " deep=\"" + this.deep + "\"" +
                (this.created > -1 ? " created=\"" + this.created + "\"" : "") +
                (this.lastFound > -1 ? " lastFound=\"" + this.lastFound + "\"" : "") +
            ">\n" +
            "\t\t\t\t\t\t<regexp><![CDATA[" + this.regexp + "]]></regexp>\n" +
            "\t\t\t\t\t\t<name><![CDATA[" + this.headerName + "]]></name>\n" +
            "\t\t\t\t\t\t<script><![CDATA[" + this.script + "]]></script>\n" +
            "\t\t\t\t\t</item>\n";
    }

    this.setRegExp(regexp);
    this.setHeaderName(headerName);
    this.setScript(script);
    this.setMerge(merge);
    this.setDeep(deep);
    this.setCreated(created ? created : Date.now());
    this.setLastFound(lastFound);
}

function YaripRedirectItem(regexp, newsubstr, deep, created, lastFound)
{
    this.regexp = "";
    this.regexpobj = null;
    this.newsubstr = "";
    this.deep = true;
    this.created = -1;
    this.lastFound = -1;

    this.getKey = function()
    {
        return this.getRegExp();
    }

    this.setRegExp = function(value)
    {
        if (!value) return;

        this.regexp = value;
        this.setRegExpObj(value);
        this.setCreated(Date.now());
    }

    this.getRegExp = function()
    {
        return this.regexp;
    }

    this.setRegExpObj = function(value)
    {
        if (value) this.regexpobj = new RegExp(value);
    }

    this.getRegExpObj = function()
    {
        return this.regexpobj;
    }

    this.setNewSubStr = function(value)
    {
        if (!value && value != "") return;

        this.newsubstr = value;
    }

    this.getNewSubStr = function()
    {
        return this.newsubstr;
    }

    this.setDeep = function(value)
    {
        this.deep = "" + value != "false";
    }

    this.getDeep = function()
    {
        return this.deep;
    }

    this.setCreated = function(value)
    {
        if (!value) return;
        value = Number(value);
        if (typeof value != "number" || value + "" == "NaN") return;
        this.created = value;
    }

    this.getCreated = function()
    {
        return this.created;
    }

    this.setLastFound = function(value)
    {
        if (!value) return;
        value = Number(value);
        if (typeof value != "number" || value + "" == "NaN") return;
        this.lastFound = value;
    }

    this.getLastFound = function()
    {
        return this.lastFound;
    }

    this.incrementLastFound = function()
    {
        this.lastFound = Date.now();
    }

    this.purge = function()
    {
        this.created = Date.now();
        this.lastFound = -1;
    }

    this.clone = function(purge)
    {
        var tmp = new this.constructor(this.regexp, this.newsubstr, this.deep, this.created, this.lastFound);
//        for (var k in this) if (this[k] && typeof this[k] != "object" && !/^(regexp|newsubstr|deep|created|lastFound)$/.test(k)) tmp[k] = this[k];
        if (purge) tmp.purge();
        return tmp;
    }

    this.merge = function(o)
    {
        if (!o) return;
        this.setRegExp(o.getRegExp() || this.getRegExp());
        this.setNewSubStr(o.getNewSubStr() || this.getNewSubStr());
        this.setDeep(o.getDeep() && this.getDeep());
        if (this.getCreated() == -1 || o.getCreated() < this.getCreated()) this.setCreated(o.getCreated());
    }

    this.compare = function(b)
    {
        var aKey = this.getKey();
        var bKey = b.getKey();
        if (aKey < bKey) return -1;
        if (aKey > bKey) return 1;
        return 0;
    }

    this.generateXml = function()
    {
        if (!this.regexp) return "";
        return "\t\t\t\t" +
            "<item" +
                " deep=\"" + this.deep + "\"" +
                (this.created > -1 ? " created=\"" + this.created + "\"" : "") +
                (this.lastFound > -1 ? " lastFound=\"" + this.lastFound + "\"" : "") +
            ">\n" +
            "\t\t\t\t\t<regexp><![CDATA[" + this.regexp + "]]></regexp>\n" +
            (this.newsubstr ? "\t\t\t\t\t<newsubstr><![CDATA[" + this.newsubstr + "]]></newsubstr>\n" : "") +
            "\t\t\t\t</item>\n";
    }

    this.setRegExp(regexp);
    this.setNewSubStr(newsubstr);
    this.setDeep(deep);
    this.setCreated(created ? created : Date.now());
    this.setLastFound(lastFound);
}

function YaripPageExtensionItem(value, doElements, doContents, doScripts, doHeaders, doRedirects, doLinks, created)
{
    this.value = "";
    this.doElements = false;
    this.doContents = true;
    this.doScripts = false;
    this.doHeaders = false;
    this.doRedirects = false;
    this.doLinks = false;
    this.created = -1;

    this.getKey = function()
    {
        return this.value;
    }

    this.setValue = function(value)
    {
        if (value) this.value = "" + value;
    }

    this.getValue = function()
    {
        return this.value;
    }

    this.setDoElements = function(value)
    {
        this.doElements = "" + value == "true";
    }

    this.getDoElements = function()
    {
        return this.doElements;
    }

    this.setDoContents = function(value)
    {
        this.doContents = "" + value != "false";
    }

    this.getDoContents = function()
    {
        return this.doContents;
    }

    this.setDoScripts = function(value)
    {
        this.doScripts = "" + value == "true";
    }

    this.getDoScripts = function()
    {
        return this.doScripts;
    }

    this.setDoHeaders = function(value)
    {
        this.doHeaders = "" + value == "true";
    }

    this.getDoHeaders = function()
    {
        return this.doHeaders;
    }

    this.setDoRedirects = function(value)
    {
        this.doRedirects = "" + value == "true";
    }

    this.getDoRedirects = function()
    {
        return this.doRedirects;
    }

    this.setDoLinks = function(value)
    {
        this.doLinks = "" + value == "true";
    }

    this.getDoLinks = function()
    {
        return this.doLinks;
    }

    this.setCreated = function(value)
    {
        if (!value) return;
        value = Number(value);
        if (typeof value != "number" || value + "" == "NaN") return;
        this.created = value;
    }

    this.getCreated = function()
    {
        return this.created;
    }


//    this.purge = function()
//    {
//        this.created = Date.now();
//    }

    this.clone = function()
    {
//        var tmp = new this.constructor(this.value, this.doElements, this.doContents, this.doScripts, this.doHeaders, this.doRedirects, this.doLinks, this.created);
        var tmp = new this.constructor(this.value, this.doElements, this.doContents, this.doScripts, this.doHeaders, this.doRedirects, this.doLinks);
//        for (var k in this) if (this[k] && typeof this[k] != "object" && !/^(value|doElements|doContents|doScripts|doHeaders|doRedirects|doLinks|created)$/.test(k)) tmp[k] = this[k];
//        if (purge) tmp.purge();
        return tmp;
    }

    this.merge = function(o)
    {
        if (!o) return;
        this.setValue(o.getValue() || this.getValue());
        this.setDoElements(o.getDoElements() || this.getDoElements());
        this.setDoContents(o.getDoContents() && this.getDoContents());
        this.setDoScripts(o.getDoScripts() || this.getDoScripts());
        this.setDoHeaders(o.getDoHeaders() || this.getDoHeaders());
        this.setDoRedirects(o.getDoRedirects() || this.getDoRedirects());
        this.setDoLinks(o.getDoLinks() || this.getDoLinks());
        if (this.getCreated() == -1 || o.getCreated() < this.getCreated()) this.setCreated(o.getCreated());
    }

    this.compare = function(b)
    {
        var aPage = this.getPage();
        var bPage = b.getPage();
        return aPage.compare(bPage);
    }

    this.generateXml = function()
    {
        if (!this.value) return "";
        return "\t\t\t\t" +
            "<item" +
                " doElements=\"" + this.doElements + "\"" +
                " doContents=\"" + this.doContents + "\"" +
                " doScripts=\"" + this.doScripts + "\"" +
                " doHeaders=\"" + this.doHeaders + "\"" +
                " doRedirects=\"" + this.doRedirects + "\"" +
                " doLinks=\"" + this.doLinks + "\"" +
                (this.created > -1 ? " created=\"" + this.created + "\"" : "") +
            ">" + this.value + "</item>\n";
    }

    this.setValue(value);
    this.setDoElements(doElements);
    this.setDoContents(doContents);
    this.setDoScripts(doScripts);
    this.setDoHeaders(doHeaders);
    this.setDoRedirects(doRedirects);
    this.setDoLinks(doLinks);
    this.setCreated(created ? created : Date.now());
}
YaripPageExtensionItem.prototype.getPage = function()
{
//    return yarip.map.getById(this.getKey());
    return Components.classes["@yarip.mozdev.org;1"].getService().wrappedJSObject.map.getById(this.getKey());
}
