
/*
Copyright 2007-2011 Kim A. Brandt <kimabrandt@gmx.de>

This file is part of yarip.

Yarip is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

Yarip is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with yarip; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

function YaripMap()
{
    this.obj = {}
    this.objId = {}
    this.tree = new YaripAATree();
    this.length = 0;
    this.version = yarip.getCurrentVersion();

    this.add = function(page)
    {
        if (!page) return;

        var oldPage = this.obj[page.getName()];
        if (oldPage) {
            this.replaceExtensionIds(this, oldPage.getId(), page.getId());
            oldPage.merge(page);
        } else {
            if (!page.getId()) page.setId(yarip.getId());
            // TODO Check if the id already exists!
            this.obj[page.getName()] = page;
            this.objId[page.getId()] = page;
            this.tree.add(page);
            this.length++;
        }
    }

    this.remove = function(page)
    {
        if (!page) return;

        var pageId = page.getId();
        if (pageId) {
            this.removeById(page.getId());
        } else {
            this.removeByName(pageName);
        }
    }

    this.get = function(pageName)
    {
        return pageName in this.obj ? this.obj[pageName] : null;
    }

    this.getById = function(value)
    {
        return value in this.objId ? this.objId[value] : null;
    }

    this.removeByName = function(pageName)
    {
        if (!pageName || !this.obj[pageName]) return;
        var page = this.obj[pageName];
        this.tree.remove(page);
        delete this.objId[page.getId()];
        delete this.obj[pageName];
        this.length--;
    }

    this.removeById = function(id)
    {
        if (!id || !this.objId[id]) return;
        var page = this.objId[id];
        this.tree.remove(page);
        delete this.obj[page.getName()];
        delete this.objId[id];
        this.length--;
    }

    this.clone = function()
    {
        var map = new this.constructor();
        for each (var page in this.obj) if (page) map.add(page.clone());
        return map;
    }

    this.replaceExtensionIds = function(map, newId, oldId)
    {
        if (!map || !newId || !oldId) return;
        var mapObj = map.obj;
        for each (var page in mapObj)
        {
            var pageExt = page.pageExtensionList;
            var extObj = pageExt.obj;
            for each (var oldItem in extObj)
            {
                if (oldItem.getValue() == oldId)
                {
                    var newItem = oldItem.clone();
                    newItem.setValue(newId);
                    pageExt.remove(oldItem);
                    pageExt.add(newItem);
                }
            }
        }
    }

    this.merge = function(map)
    {
        if (!map) return;
        var mapObj = map.obj;
        for (var p in mapObj)
        {
            var page = map.obj[p];
            if (this.obj[p]) {
                this.replaceExtensionIds(map, this.obj[p].getId(), page.getId());
            } else if (this.objId[page.getId()]) {
                var newId = "" + yarip.getId();
                this.replaceExtensionIds(map, newId, "" + page.getId());
                page.setId(newId);
            }
        }
        for each (var page in mapObj) {
            this.add(page);
        }
    }

    this.purge = function()
    {
        for each (var page in this.obj) if (page) page.purge();
    }

    this.generateXml = function(exporting)
    {
        var r = "<?xml version=\"1.0\" encoding=\"" + yarip.CHARSET + "\"?>\n";
        if (!exporting) {
            r += "\n<!--\nDo not edit this file.\n\nIf you make changes to this file while the application is running,\nthe changes will be overwritten when the application exits.\n-->\n\n";
        }
        r += "<yarip version=\"" + this.version + "\">\n";
        this.tree.traverse(function(node) { if (node.value) r += node.value.generateXml(); });
        r += "</yarip>";
        return r;
    }
}
