
/*
Copyright 2007-2011 Kim A. Brandt <kimabrandt@gmx.de>

This file is part of yarip.

Yarip is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

Yarip is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with yarip; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

function YaripOverlay()
{
    this.enabledObserver = null;
    this.flickerObserver = null;
    this.modeObserver = null;
    this.alwaysUseIndexObserver = null;
    this.elementsInContextObserver = null;
    this.purgeInnerHTMLObserver = null;
    this.exclusiveOnCreationObserver = null;
    this.templatesObserver = null;
    this.schemesObserver = null;
    this.privateObserver = null;

    this.appcontent = null;
    this.gBrowser = null;
    this.contextMenu = null;
    this.yaripMenupopup = null;
    this.yaripMonitorMenupopup = null;
    this.tabContainer = null;
    this.stringbundle = null;

    this.doc = null;
    this.loader = null;
    this.handler = null;

    this.active = false;

    this.timeout = null;
    this.statusTimeout = null;

    this.getAppcontent = null;
    this.getBrowser = null;
    this.getContextMenu = null;
    this.getContextMenuId = null;
    this.getTabContainer = null;

    this.start = function(event)
    {
        if (event.button != 0) return;

        var doc = content.document;
        if (this.active) {
            this.stop(doc, false);
            return;
        }

        if (!yarip.enabled) {
            this.toggleEnabled(true);
            return;
        }

        if (!doc) return;

        if (!this.handler.start(doc)) {
            this.stop(doc, false);
            return;
        }

        this.active = true;
        this.doc = doc;
        this.setYaripStatus(doc, "started");
    }

    this.stop = function(doc, noStateChange, noSort)
    {
        if (!this.active || !this.handler) return false;

        this.hideStatus();

        doc = doc ? doc : this.doc;
        if (this.handler.stop()) {
            if (doc) {
                var location = yarip.getLocationFromLocation(doc.location);
                yarip.reloadPage(yarip.getPageName(location));
            }
        }
        this.active = false;

        if (!doc) return true;

        if (!noStateChange) this.setYaripStatus(doc, "stopped");
        this.doc = null;

        return true;
    }

    this.whitelistXPath = function(doc, node)
    {
        this.toggleEnabled(true);

        if (!doc || !node) return;

        var location = yarip.getLocationFromLocation(doc.location);
        var address = yarip.getFirstAddress(location.asciiHref);
        if (!address) {
            address = yarip.getPageName(location);
            if (!address) return;
        }

        var xValue = yarip.calculateXPath(node);
        if (!xValue) return;

        var obj = {
            pageName: address,
            element: new YaripElementWhitelistItem(xValue + "/descendant-or-self::*")
        }

        window.openDialog("chrome://yarip/content/whitelistelementdialog.xul", "whitelistelementdialog", "chrome,modal,resizable", doc, obj);

        if (!obj.pageName || obj.pageName == "") return;
        if (!obj.element) return;

//        yarip.whitelistXPath(doc, obj.pageName, obj.element.getXPath(), true, obj.element, true);
        yarip.whitelistXPath(doc, obj.pageName, obj.element.getXPath(), true, obj.element.getForce(), obj.element, true);
    }

    this.blacklistXPath = function(doc, node)
    {
        this.toggleEnabled(true);

        if (!doc || !node) return;

        var location = yarip.getLocationFromLocation(doc.location);
        var address = yarip.getFirstAddress(location.asciiHref);
        if (!address) {
            address = yarip.getPageName(location);
            if (!address) return;
        }

        var xValue = yarip.calculateXPath(node);
        if (!xValue) return;

        var obj = {
            pageName: address,
            element: new YaripElementBlacklistItem(xValue)
        }

        window.openDialog("chrome://yarip/content/blacklistelementdialog.xul", "blacklistelementdialog", "chrome,modal,resizable", doc, obj);

        if (!obj.element) return;
        if (!obj.pageName || obj.pageName == "") return;

        yarip.blacklistXPath(doc, obj.pageName, obj.element.getXPath(), true, obj.element.getForce(), obj.element, true);
    }

    this.whitelistTemporarily = function(doc, node)
    {
        this.toggleEnabled(true);

        if (!doc || !node) return;

        var xValue = yarip.calculateXPath(node);
        if (!xValue) return;

        yarip.whitelistXPath(doc, null, xValue + "/descendant-or-self::*");
        yarip.removeAllExceptWhitelisted(doc);
    }

    this.removeTemporarily = function(doc, node)
    {
        this.toggleEnabled(true);

        if (!doc || !node) return;

        var xValue = yarip.calculateXPath(node);
        if (!xValue) return;

        yarip.blacklistXPath(doc, null, xValue, false, true);
    }

    this.styleElement = function(doc, node, attrName)
    {
        this.toggleEnabled(true);

        if (!doc || !node) return;

        var location = yarip.getLocationFromLocation(doc.location);
        var address = yarip.getFirstAddress(location.asciiHref);
        if (!address) {
        address = yarip.getPageName(location);
        if (!address) return;
      }

        var xValue = yarip.calculateXPath(node);
        if (!xValue) return;

        var attrValue = "";
        if (attrName)
        {
            attrValue = node.getAttribute(attrName);
            if (attrName == "style") attrValue = attrValue.replace(yarip.endOutlineRegExp, "");
        }

        var obj = {
            pageName: address,
            element: new YaripElementAttributeItem(xValue),
            attrName: attrName ? attrName : "",
            attrValue: attrValue,
            node: node
        }

        window.openDialog("chrome://yarip/content/styledialog.xul", "styledialog", "chrome,modal,resizable", doc, obj);

        if (!obj.pageName || obj.pageName == "") return;
        if (!obj.element) return;
        if (!obj.attrName || obj.attrName == "") return;

        obj.element.setName(obj.attrName);
        obj.element.setValue(obj.attrValue);

        yarip.styleElement(doc, obj.pageName, obj.element.getXPath(), obj.attrName, obj.attrValue, true, obj.element, true);
    }

    this.scriptElement = function(doc, node)
    {
        this.toggleEnabled(true);

        if (!doc || !node) return;

        var location = yarip.getLocationFromLocation(doc.location);
        var address = yarip.getFirstAddress(location.asciiHref);
        if (!address) {
        address = yarip.getPageName(location);
        if (!address) return;
      }

        var xValue = yarip.calculateXPath(node);
        if (!xValue) return;

        var obj = {
            element: new YaripElementScriptItem(xValue,
                    "function (array) {\n" +
                    "    // Do something with the elements in the array!\n" +
                    "}"),
            pageName: address
        }

        window.openDialog("chrome://yarip/content/scriptdialog.xul", "scriptdialog", "chrome,modal,resizable", doc, obj);

        if (!obj.pageName || obj.pageName == "") return;
        if (!obj.element) return;

        yarip.scriptElement(doc, obj.pageName, obj.element.getXPath(), true, obj.element);
    }

    this.copyXPath = function(node)
    {
        if (!node) return;

        var xValue = yarip.calculateXPath(node);
        if (!xValue) return;

        var ch = Components.classes["@mozilla.org/widget/clipboardhelper;1"].getService(Components.interfaces.nsIClipboardHelper);
        ch.copyString(xValue);
    }

    this.startHighlighting = function(node)
    {
        if (!node) return;

        this.stopHighlighting();
        this.handler.doc = node.ownerDocument;
        this.handler.element = node;
        this.handler.select();
    }

    this.stopHighlighting = function()
    {
        this.handler.doc = null;
        this.handler.reset();
    }

    this.createPage = function(location)
    {
        this.stop();

        yarip.resetUndo();

        location = yarip.getLocationFromLocation(location ? location : content.document.location);
        var pageName = yarip.getFirstAddress(location.asciiHref);
        if (!pageName) {
            pageName = yarip.getPageName(location);
            if (!pageName) return;
        }

        var obj = {
            pageName: pageName,
            pageLocation: location
//            contentLocation: null
        }

        window.openDialog("chrome://yarip/content/createpagedialog.xul", "createpagedialog", "chrome,modal,resizable", obj);

        if (!obj.pageName) return;

        var page = yarip.createPage(obj.pageLocation, obj.pageName);
        this.managePages(page.getName());
    }

    this.monitorContent = function()
    {
        this.stop();
        yaripMonitor.show();
    }

    this.managePages = function(pageName)
    {
        this.stop();

        yarip.resetUndo();

        if (!pageName) {
            var location = yarip.getLocationFromLocation(content.document.location);
            pageName = yarip.getFirstAddress(location.asciiHref);
        }

        if (!yarip.reloadPage(pageName, true, true)) {
            window.openDialog("chrome://yarip/content/pagedialog.xul", "pagedialog", "chrome,resizable", pageName);
        }
    }

    this.toggleEnabled = function(enabled, force)
    {
        this.stop();
        this.enabledObserver.disable();
        var enabledChanged = yarip.toggleEnabled(enabled) || force;
        this.enabledObserver.enable();
        if (enabledChanged) {
            document.getElementById("yarip-enabled-menuitem").setAttribute("checked", "" + yarip.enabled);
            this.setYaripStatus(content.document, yarip.enabled ? "enabled" : "disabled");
        }
    }

    this.updateEnabled = function()
    {
        this.toggleEnabled(yarip.getValue(yarip.PREF_ENABLED, true, yarip.DATA_TYPE_BOOLEAN), true);
    }

    this.toggleNoFlicker = function(noFlicker, force)
    {
        this.flickerObserver.disable();
        if (!yarip.toggleNoFlicker(noFlicker) && !force)
        {
            this.flickerObserver.enable();
            return;
        }
        this.flickerObserver.enable();

        document.getElementById("yarip-flicker-menuitem").setAttribute("checked", "" + yarip.noFlicker);
    }

    this.updateNoFlicker = function()
    {
        this.toggleNoFlicker(yarip.getValue(yarip.PREF_FLICKER, false, yarip.DATA_TYPE_BOOLEAN), true);
    }

    this.setMode = function(mode, clear, force, update)
    {
        var address = yarip.getPageName(content.document.location);

        this.modeObserver.disable();

        if (!update)
        {
            if (!yarip.setMode(mode) && !force)
            {
                this.modeObserver.enable();
                return;
            }
            this.modeObserver.enable();
        }
        else this.modeObserver.enable();

        if (clear)
        {
            document.getElementById("yarip-page-menuitem").setAttribute("checked", "false");
            document.getElementById("yarip-fqdn-menuitem").setAttribute("checked", "false");
            document.getElementById("yarip-sld-menuitem").setAttribute("checked", "false");
        }

        this.stop();

        switch (mode)
        {
            case yarip.MODE_PAGE:
                document.getElementById("yarip-page-menuitem").setAttribute("checked", "true");
                break;

            case yarip.MODE_FQDN:
                document.getElementById("yarip-fqdn-menuitem").setAttribute("checked", "true");
                break;

            case yarip.MODE_SLD:
                document.getElementById("yarip-sld-menuitem").setAttribute("checked", "true");
                break;

            default:
                this.setMode(yarip.MODE_FQDN, true);
        }
    }

    this.updateMode = function()
    {
        this.setMode(yarip.getValue(yarip.PREF_MODE, yarip.MODE_FQDN, yarip.DATA_TYPE_INTEGER), true, true, true);
    }

    this.handleEvent = function(event)
    {
        switch(event.type) {
        case "load":
            yaripMonitor.load();

            this.loader = new YaripLoader();
            this.handler = new YaripHandler(null, function() { yaripOverlay.stop(); }, function(status) { yaripOverlay.setStatus(status); });

            this.enabledObserver = new YaripObserver(
                yarip.PREF_ENABLED,
                function() { yaripOverlay.updateEnabled(); }
            );
            this.flickerObserver = new YaripObserver(
                yarip.PREF_FLICKER,
                function() { yaripOverlay.updateNoFlicker(); }
            );
            this.modeObserver = new YaripObserver(
                yarip.PREF_MODE,
                function() { yaripOverlay.updateMode(); }
            );
            this.alwaysUseIndexObserver = new YaripObserver(
                yarip.PREF_INDEX,
                function() {
                    yarip.setAlwaysUseIndex(yarip.getValue(yarip.PREF_INDEX, false, yarip.DATA_TYPE_BOOLEAN));
                }
            );
            this.elementsInContextObserver = new YaripObserver(
                yarip.PREF_ELEMENTS,
                function() {
                    yarip.setElementsInContext(yarip.getValue(yarip.PREF_ELEMENTS, 4, yarip.DATA_TYPE_INTEGER));
                }
            );
            this.purgeInnerHTMLObserver = new YaripObserver(
                yarip.PREF_PURGE,
                function() {
                    yarip.setPurgeInnerHTML(yarip.getValue(yarip.PREF_PURGE, false, yarip.DATA_TYPE_BOOLEAN));
                }
            );
            this.exclusiveOnCreationObserver = new YaripObserver(
                yarip.PREF_EXCLUSIVE,
                function() {
                    yarip.setExclusiveOnCreation(yarip.getValue(yarip.PREF_EXCLUSIVE, false, yarip.DATA_TYPE_BOOLEAN));
                }
            );
            this.templatesObserver = new YaripObserver(
                yarip.PREF_TEMPLATES,
                function() {
                    yarip.setTemplates(yarip.getValue(yarip.PREF_TEMPLATES, "", yarip.DATA_TYPE_STRING));
                }
            );
            this.schemesObserver = new YaripObserver(
                yarip.PREF_SCHEMES,
                function() {
                    yarip.setSchemes(yarip.getValue(yarip.PREF_SCHEMES, "", yarip.DATA_TYPE_STRING));
                }
            );
            this.privateObserver = new YaripObserver(
                yarip.PREF_PRIVATE,
                function() {
                    yarip.setPrivateBrowsing(yarip.getValue(yarip.PREF_PRIVATE, false, yarip.DATA_TYPE_BOOLEAN));
                }
            );

            var y = document.getElementById("yarip");
            this.getAppcontent = function() new Function(y.getAttribute("getAppcontent")).call(window);
            this.getBrowser = function() new Function(y.getAttribute("getBrowser")).call(window);
            this.getContextMenuId = function() new Function(y.getAttribute("getContextMenuId")).call(window);
            this.getContextMenu = function() new Function(y.getAttribute("getContextMenu")).call(window);
            this.getTabContainer = function() new Function(y.getAttribute("getTabContainer")).call(window);

            this.appcontent = this.getAppcontent();
            this.gBrowser = this.getBrowser();
            this.contextMenu = this.getContextMenu();
            this.contextMenuId = this.getContextMenuId();
            this.yaripMenupopup = document.getElementById("yarip-menupopup");
            this.yaripMonitorMenupopup = document.getElementById("yarip-monitor-menupopup");
            this.tabContainer = this.getTabContainer();
            this.stringbundle = document.getElementById("yarip-overlay-stringbundle");
            this.updateEnabled();
            this.updateNoFlicker();
            this.updateMode();

            yarip.injectCSS("chrome://yarip/skin/blacklist.css", true);

            if (this.appcontent) this.appcontent.addEventListener("DOMContentLoaded", this, false);
            if (this.contextMenu) this.contextMenu.addEventListener("popupshowing", this, false);
            if (this.yaripMonitorMenupopup) this.yaripMonitorMenupopup.addEventListener("popupshowing", yaripMonitor, false);
            if (this.tabContainer) this.tabContainer.addEventListener("select", this, false);
            window.removeEventListener("load", this, false);
            break;
        case "unload":
            this.enabledObserver.unregister();
            this.flickerObserver.unregister();
            this.modeObserver.unregister();
            this.alwaysUseIndexObserver.unregister();
            this.elementsInContextObserver.unregister();
            this.purgeInnerHTMLObserver.unregister();
            this.exclusiveOnCreationObserver.unregister();
            this.templatesObserver.unregister();
            this.schemesObserver.unregister();
            this.privateObserver.unregister();

            yaripMonitor.unload();

            if (this.appcontent) this.appcontent.removeEventListener("DOMContentLoaded", this, false);
            if (this.contextMenu) this.contextMenu.removeEventListener("popupshowing", this, false);
            if (this.tabContainer) this.tabContainer.removeEventListener("select", this, false);
            window.removeEventListener("unload", this, false);
            break;
        case "DOMContentLoaded":
            this.stop();
            yarip.resetUndo();

            var doc = event.originalTarget;
            if (!doc || !doc.body) break;
            this.doDOMContentLoaded(doc, this);
            doc.body.setAttribute("status", "whitelisted");
            doc.body.addEventListener("DOMNodeInserted", this, false);
            break;
        case "DOMNodeInserted":
            if (/firebugLayoutBox|firebugHighlight/.test(event.originalTarget.getAttribute("class"))) break;

            if (this.timeout) clearTimeout(this.timeout);
            this.timeout = setTimeout(this.doDOMContentLoaded, 1000, event.originalTarget.ownerDocument, this);
            break;
        case "DOMMenuItemActive":
            var menu = event.target;
            this.startHighlighting(menu.attribute ? menu.attribute : menu.node);
            break;
        case "DOMMenuItemInactive":
            this.stopHighlighting();
            this.hideStatusAfterTimeout();
            break;
        case "popupshowing":
            this.stop();
            if (event.target.getAttribute("id") != this.contextMenuId) return;

            // YARIP-MENU
            var menu = this.yaripMenupopup.firstChild;
            while (menu && menu.getAttribute("id") != "yarip-undo-menu-sep")
            {
                var next = menu.nextSibling;
                menu.removeEventListener("DOMMenuItemActive", this, false);
                menu.removeEventListener("DOMMenuItemInactive", this, false);
                this.yaripMenupopup.removeChild(menu);
                menu = next;
            }
            var node = document.popupNode;
            var doc = node.ownerDocument;
            var hasMenus = false;
            var location = yarip.getLocationFromLocation(doc.location);
            if (yarip.getPageName(location))
            {
                var undoMenuSep = document.getElementById("yarip-undo-menu-sep");
                var undoMenu = document.getElementById("yarip-undo-menupopup");
                undoMenuSep.hidden = false;

                // ELEMENT-MENUES
                var tmpNode = node;
                for (var i = 0; i < yarip.elementsInContext && tmpNode && tmpNode.localName && !/^(html|body|frameset|frame)$/i.test(tmpNode.localName); i++)
                {
                    menu = this.createMenu(tmpNode);
                    if (menu) {
                        this.yaripMenupopup.insertBefore(menu, undoMenuSep);
                        hasMenus = true;
                    }
                    tmpNode = tmpNode.parentNode;
                }

                // FRAME-MENU
                var inFrame = gContextMenu.inFrame;
                if (inFrame)
                {
                    var frame = node.ownerDocument.defaultView.frameElement;
                    var frameMenu = this.createMenu(frame, true);
                    if (!yarip.getPageName(frame.ownerDocument.location)) frameMenu.setAttribute("disabled", true);
                    this.yaripMenupopup.insertBefore(frameMenu, undoMenuSep);
                }
                hasMenus = hasMenus || inFrame;

                // UNDO-MENU
                var menuitem = undoMenu.firstChild;
                while (menuitem && menuitem.localName != "menuseparator")
                {
                    var next = menuitem.nextSibling;
                    undoMenu.removeChild(menuitem);
                    menuitem = next;
                }
                var found = false;
                for each (var obj in yarip.undoObj)
                {
                    if (!obj || obj.document != doc) continue;

                    var menuitem = document.createElement("menuitem");
                    menuitem.setAttribute("label", this.stringbundle.getFormattedString("undo-" + obj.type, [obj.text]));
                    menuitem.obj = obj;
                    menuitem.setAttribute("oncommand", "yaripOverlay.undo(this.obj)");
                    undoMenu.insertBefore(menuitem, undoMenu.firstChild);
                    found = true;
                }
                document.getElementById("yarip-undo-menu-sep").hidden = !hasMenus;
                document.getElementById("yarip-undo-menu").hidden = !found;
            } else {
                document.getElementById("yarip-undo-menu-sep").hidden = true;
            }
            break;
        case "select":
            if (this.gBrowser) this.setYaripStatus(this.gBrowser.contentWindow.document);
            break;
        case "keypress":
//            javascript:alert('y'.charCodeAt())
            if (event.ctrlKey && event.keyCode == 0) {
                if (event.altKey) {
                    switch (event.charCode) {
                    case 109: // <Ctrl> + <Alt> + <m>
                        this.managePages();
                        event.stopPropagation();
                        break;
                    }
                } else {
                    switch (event.charCode) {
                    case 121: // <Ctrl> + <y>
                        yaripMonitor.toggleShow();
                        event.stopPropagation();
                        break;
                    }
                }
            }
            break;
        }
    }

    this.doDOMContentLoaded = function(doc, overlay)
    {
        if (!yarip.enabled) return;
        if (!doc || !doc.body || !doc.location) return;
        if (!/^https?:$/.test(doc.location.protocol)) return;
        if (!/^(text\/html|application\/xhtml\+xml)$/.test(doc.contentType)) return;

        if (overlay.loader.load(doc, true)) overlay.setYaripStatus(doc, "found");
        else overlay.setYaripStatus(doc);
    }

    this.createMenu = function(node, isFrame)
    {
        var menu = document.createElement("menu");
        menu.node = node;
        menu.setAttribute("label", node.localName.toUpperCase());
        menu.addEventListener("DOMMenuItemActive", this, false);
        menu.addEventListener("DOMMenuItemInactive", this, false);
        var menupopup = document.createElement("menupopup");

//        if (!isFrame)
//        {
            // WHITELIST ELEMENT
            var weMenuitem = document.createElement("menuitem");
            weMenuitem.node = node;
            weMenuitem.setAttribute("label", this.stringbundle.getString("menuitemElementWhitelist"));
            weMenuitem.setAttribute("oncommand", "yaripOverlay.whitelistXPath(this.node.ownerDocument, this.node)");
            weMenuitem.setAttribute("class", "whitelist");
            menupopup.appendChild(weMenuitem);

            // BLACKLIST ELEMENT
            var beMenuitem = document.createElement("menuitem");
            beMenuitem.node = node;
            beMenuitem.setAttribute("label", this.stringbundle.getString("menuitemElementBlacklist"));
            beMenuitem.setAttribute("oncommand", "yaripOverlay.blacklistXPath(this.node.ownerDocument, this.node)");
            beMenuitem.setAttribute("class", "blacklist");
            menupopup.appendChild(beMenuitem);
//        }

        // BLACKLIST ATTRIBUTE
        if (node.hasAttributes())
        {
            var found = false;
            var baMenu = document.createElement("menu");
            baMenu.setAttribute("label", this.stringbundle.getString("menuAttributeBlacklist"));
            var baMenupopup = document.createElement("menupopup");
            if (node.hasAttributes()) {
                for (var i = 0; i < node.attributes.length; i++)
                {
                    var attribute = node.attributes[i];
                    if (!attribute || attribute.value == "") continue;

                    switch (attribute.name)
                    {
                        case "status":
                            if (/^(whitelisted|blacklisted)$/.test(attribute.value)) continue;

                        case "style":
                            if (yarip.outlineRegExp.test(attribute.value)) continue;
                    }

                    var baMenuitem = document.createElement("menuitem");
                    baMenuitem.node = node;
                    baMenuitem.setAttribute("label", attribute.name.toLowerCase());
                    baMenuitem.attribute = attribute;
                    if (isFrame) baMenuitem.setAttribute("oncommand", "yaripOverlay.blacklistXPath(this.node.ownerDocument.defaultView.frameElement.ownerDocument, this.attribute)");
                    else baMenuitem.setAttribute("oncommand", "yaripOverlay.blacklistXPath(this.node.ownerDocument, this.attribute)");
                    baMenupopup.appendChild(baMenuitem);
                    found = true;
                }
            }

            if (found) {
                baMenu.appendChild(baMenupopup);
                menupopup.appendChild(baMenu);
            }
        }

        // STYLE ELEMENT
        var seMenu = document.createElement("menu");
        seMenu.setAttribute("label", this.stringbundle.getString("menuitemStyleElement"));
        var seMenuitem = document.createElement("menuitem");
        seMenuitem.node = node;
        seMenuitem.setAttribute("label", this.stringbundle.getString("menuitemUserDefined"));
        if (isFrame) seMenuitem.setAttribute("oncommand", "yaripOverlay.styleElement(this.node.ownerDocument.defaultView.frameElement.ownerDocument, this.node)");
        else seMenuitem.setAttribute("oncommand", "yaripOverlay.styleElement(this.node.ownerDocument, this.node)");
        var seMenupopup = document.createElement("menupopup");
        seMenupopup.appendChild(seMenuitem);
        if (node.hasAttributes()) {
            var seMenuseparator = document.createElement("menuseparator");
            seMenupopup.insertBefore(seMenuseparator, seMenuitem);
            for (var i = 0; i < node.attributes.length; i++)
            {
                var attribute = node.attributes[i];
                if (!attribute || attribute.value == "") continue;

                switch (attribute.name) {
                    case "status": if (/^(whitelisted|blacklisted)$/.test(attribute.value)) continue;
                    case "style": if (yarip.outlineRegExp.test(attribute.value)) continue;
                }

                var seMenuitem = document.createElement("menuitem");
                seMenuitem.node = node;
                seMenuitem.setAttribute("label", attribute.name.toLowerCase());
                seMenuitem.attribute = attribute;
                if (isFrame) seMenuitem.setAttribute("oncommand", "yaripOverlay.styleElement(this.node.ownerDocument.defaultView.frameElement.ownerDocument, this.node, '" + attribute.name.toLowerCase() + "')");
                else seMenuitem.setAttribute("oncommand", "yaripOverlay.styleElement(this.node.ownerDocument, this.node, '" + attribute.name.toLowerCase() + "')");
                seMenupopup.insertBefore(seMenuitem, seMenuseparator);
            }
        }
        seMenu.appendChild(seMenupopup);
        menupopup.appendChild(seMenu);

        // SCRIPT ELEMENT
        var sceMenuitem = document.createElement("menuitem");
        sceMenuitem.node = node;
        sceMenuitem.setAttribute("label", this.stringbundle.getString("menuitemScriptElement"));
        sceMenuitem.setAttribute("oncommand", "yaripOverlay.scriptElement(this.node.ownerDocument, this.node)");
//        sceMenuitem.setAttribute("class", "script");
        menupopup.appendChild(sceMenuitem);

//        if (!isFrame)
//        {
            // TEMPORARILY WHITELIST
            var wtMenuitem = document.createElement("menuitem");
            wtMenuitem.node = node;
            wtMenuitem.setAttribute("label", this.stringbundle.getString("menuitemKeepTemporarily"));
            wtMenuitem.setAttribute("oncommand", "yaripOverlay.whitelistTemporarily(this.node.ownerDocument, this.node)");
            wtMenuitem.setAttribute("class", "whitelist temporary");
            menupopup.appendChild(wtMenuitem);

            // TEMPORARILY BLACKLIST
            var rtMenuitem = document.createElement("menuitem");
            rtMenuitem.node = node;
            rtMenuitem.setAttribute("label", this.stringbundle.getString("menuitemRemoveTemporarily"));
            rtMenuitem.setAttribute("oncommand", "yaripOverlay.removeTemporarily(this.node.ownerDocument, this.node)");
            rtMenuitem.setAttribute("class", "blacklist temporary");
            menupopup.appendChild(rtMenuitem);
//        }

        // COPY XPATH
        var cxMenuitem = document.createElement("menuitem");
        cxMenuitem.node = node;
        cxMenuitem.setAttribute("label", this.stringbundle.getString("menuitemCopyXPath"));
        cxMenuitem.setAttribute("oncommand", "yaripOverlay.copyXPath(this.node)");
        cxMenuitem.setAttribute("class", "copyXPath");
        menupopup.appendChild(cxMenuitem);

        menu.appendChild(menupopup);
        return menu;
    }

    this.undoAll = function(doc)
    {
        if (!doc) return;

        var obj = null;
        for each (obj in yarip.undoObj) {
            if (obj && obj.document === doc) this.undo(obj, true);
        }

        if (obj) yarip.reloadPage(obj.address);
    }

    this.undo = function(obj, noReload)
    {
        if (!obj) return;

        yarip.resetOnAddress(obj);
        if (!noReload) yarip.reloadPage(obj.address);
    }

    this.setStatus = function(status)
    {
        clearTimeout(this.statusTimeout);
        document.getElementById("yarip-status-label").value = status;
        document.getElementById("yarip-status").hidden = false;
    }

    this.hideStatus = function()
    {
        document.getElementById("yarip-status").hidden = true;
    }

    this.hideStatusAfterTimeout = function()
    {
        clearTimeout(this.statusTimeout);
        this.statusTimeout = setTimeout(this.hideStatus, 100);
    }

    this.setYaripStatus = function(doc, status)
    {
        if (!doc) return;

        if (yarip.enabled)
        {
            if (!status)
            {
                status = doc.defaultView && doc.defaultView.yaripStatus ? doc.defaultView.yaripStatus : null;
                switch (status)
                {
                    case "enabled":
                    case "stopped":
                    case "found":
                        break;

                    case "disabled":
                    case "started":
                        if (this.active) break;

                    default:
                        status = "enabled";
                }
            }
        } else {
            status = "disabled";
        }

        if (doc.defaultView) {
            doc.defaultView.yaripStatus = status;
        }

        if (doc == this.gBrowser.contentWindow.document)
        {
            var toolbarbutton = document.getElementById("yarip-toolbarbutton");
            if (toolbarbutton) {
                document.getElementById("yarip-toolbarbutton").setAttribute("status", status);
                document.getElementById("yarip-toolbarbutton").setAttribute("tooltiptext", this.stringbundle.getString("status-" + status));
            }
        }
    }
}

