
/*
Copyright 2007-2011 Kim A. Brandt <kimabrandt@gmx.de>

This file is part of yarip.

Yarip is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

Yarip is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with yarip; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

function YaripPage(id, name, elementWhitelist, elementBlacklist, elementAttributeList, elementScriptList, contentWhitelist, contentBlacklist, contentRequestHeaderList, contentResponseHeaderList, contentRedirectList, pageStyleList, pageScriptList, pageRequestHeaderList, pageResponseHeaderList, pageRedirectList, pageExtensionList)
{
    this.id = null;
    this.name = "";
    this.elementWhitelist = elementWhitelist ? elementWhitelist : new YaripElementWhitelist("whitelist");
    this.elementBlacklist = elementBlacklist ? elementBlacklist : new YaripElementBlacklist("blacklist");
    this.elementAttributeList = elementAttributeList ? elementAttributeList : new YaripElementAttributeList("attribute");
    this.elementScriptList = elementScriptList ? elementScriptList : new YaripElementScriptList("script");
    this.contentWhitelist = contentWhitelist ? contentWhitelist : new YaripContentWhitelist("whitelist");
    this.contentBlacklist = contentBlacklist ? contentBlacklist : new YaripContentBlacklist("blacklist");
    this.contentRequestHeaderList = contentRequestHeaderList ? contentRequestHeaderList : new YaripHeaderList("request");
    this.contentResponseHeaderList = contentResponseHeaderList ? contentResponseHeaderList : new YaripHeaderList("response");
    this.contentRedirectList = contentRedirectList ? contentRedirectList : new YaripRedirectList("redirect");
    this.pageStyleList = pageStyleList ? pageStyleList : new YaripPageStyleList("style");
    this.pageScriptList = pageScriptList ? pageScriptList : new YaripPageScriptList("script");
    this.pageRequestHeaderList = pageRequestHeaderList ? pageRequestHeaderList : new YaripHeaderList("request");
    this.pageResponseHeaderList = pageResponseHeaderList ? pageResponseHeaderList : new YaripHeaderList("response");
    this.pageRedirectList = pageRedirectList ? pageRedirectList : new YaripRedirectList("redirect");
    this.pageExtensionList = pageExtensionList ? pageExtensionList : new YaripPageExtensionList("extension");
    this.pageExtendedByList = new YaripPageExtendedByList();
    this.temporary = false;

    this.getKey = function()
    {
        return this.getId();
    }

    this.setId = function(value)
    {
        if (!value) return;

        this.id = "" + value;
        this.pageExtensionList.setId(this.id);
        this.pageExtendedByList.setId(this.id);
    }

    this.getId = function()
    {
        return this.id;
    }

    this.setName = function(value, init)
    {
        if (value) {
            this.name = "" + value;
            if (init) this.init();
        }
    }

    this.getName = function()
    {
        return this.name;
    }

    this.isEmpty = function()
    {
        return this.elementWhitelist.isEmpty()
            && this.elementBlacklist.isEmpty()
            && this.elementAttributeList.isEmpty()
            && this.elementScriptList.isEmpty()
            && this.contentWhitelist.isEmpty()
            && this.contentBlacklist.isEmpty()
            && this.contentRequestHeaderList.isEmpty()
            && this.contentResponseHeaderList.isEmpty()
            && this.contentRedirectList.isEmpty()
            && this.pageStyleList.isEmpty()
            && this.pageScriptList.isEmpty()
            && this.pageRequestHeaderList.isEmpty()
            && this.pageResponseHeaderList.isEmpty()
            && this.pageRedirectList.isEmpty()
            && this.pageExtensionList.isEmpty();
    }

    this.setTemporary = function(value)
    {
        this.temporary = "" + value == "true";
    }

    this.getTemporary = function()
    {
        return this.temporary;
    }

    this.clone = function(purge, pageName)
    {
        var tmp = new this.constructor(
                purge ? null : this.id,
                pageName ? pageName : this.name,
                this.elementWhitelist.clone(purge),
                this.elementBlacklist.clone(purge),
                this.elementAttributeList.clone(purge),
                this.elementScriptList.clone(purge),
                this.contentWhitelist.clone(purge),
                this.contentBlacklist.clone(purge),
                this.contentRequestHeaderList.clone(purge),
                this.contentResponseHeaderList.clone(purge),
                this.contentRedirectList.clone(purge),
                this.pageStyleList.clone(),
                this.pageScriptList.clone(),
                this.pageRequestHeaderList.clone(purge),
                this.pageResponseHeaderList.clone(purge),
                this.pageRedirectList.clone(purge),
                this.pageExtensionList.clone(purge));
//        for (var k in this) {
//            if (this[k] && typeof(this[k]) != "object" && !/^(id|name)$/.test(k)) {
//                tmp[k] = this[k];
//            }
//        }
        return tmp;
    }

    this.merge = function(page)
    {
        if (!page) return;
        this.elementWhitelist.merge(page.elementWhitelist);
        this.elementBlacklist.merge(page.elementBlacklist);
        this.elementAttributeList.merge(page.elementAttributeList);
        this.elementScriptList.merge(page.elementScriptList);
        this.contentWhitelist.merge(page.contentWhitelist);
        this.contentBlacklist.merge(page.contentBlacklist);
        this.contentRequestHeaderList.merge(page.contentRequestHeaderList);
        this.contentResponseHeaderList.merge(page.contentResponseHeaderList);
        this.contentRedirectList.merge(page.contentRedirectList);
        this.pageStyleList.merge(page.pageStyleList);
        this.pageScriptList.merge(page.pageScriptList);
        this.pageRequestHeaderList.merge(page.pageRequestHeaderList);
        this.pageResponseHeaderList.merge(page.pageResponseHeaderList);
        this.pageRedirectList.merge(page.pageRedirectList);
        this.pageExtensionList.merge(page.pageExtensionList);
        this.pageExtensionList.removeByKey(this.getId());
        this.setTemporary(this.getTemporary() || page.getTemporary());
    }

    this.subtract = function(page)
    {
        if (!page) return;
        this.elementWhitelist.subtract(page.elementWhitelist);
        this.elementBlacklist.subtract(page.elementBlacklist);
        this.elementAttributeList.subtract(page.elementAttributeList);
        this.elementScriptList.subtract(page.elementScriptList);
        this.contentWhitelist.subtract(page.contentWhitelist);
        this.contentBlacklist.subtract(page.contentBlacklist);
        this.contentRequestHeaderList.subtract(page.contentRequestHeaderList);
        this.contentResponseHeaderList.subtract(page.contentResponseHeaderList);
        this.contentRedirectList.subtract(page.contentRedirectList);
        this.pageStyleList.subtract(page.pageStyleList);
        this.pageScriptList.subtract(page.pageScriptList);
        this.pageRequestHeaderList.subtract(page.pageRequestHeaderList);
        this.pageResponseHeaderList.subtract(page.pageResponseHeaderList);
        this.pageRedirectList.subtract(page.pageRedirectList);
        this.pageExtensionList.subtract(page.pageExtensionList);
        this.setTemporary(false);
    }

    this.purge = function()
    {
        this.elementWhitelist.purge();
        this.elementBlacklist.purge();
        this.elementAttributeList.purge();
        this.elementScriptList.purge();
        this.contentWhitelist.purge();
        this.contentBlacklist.purge();
        this.contentRequestHeaderList.purge();
        this.contentResponseHeaderList.purge();
        this.contentRedirectList.purge();
        this.pageRequestHeaderList.purge();
        this.pageResponseHeaderList.purge();
        this.pageRedirectList.purge();
        this.setTemporary(false);
    }

    var IP = 0;
    var SIMPLE = 1;
    var DOMAIN = 2;
    var UNKNOWN = 3;
    this.obj = {
        type: UNKNOWN,
        protocol: "",
        ipArr: [],
        tldArr: [],
        domainArr: []
    };
    this.init = function()
    {
        if (!this.getName()) return;

        var matches = this.getName().match(yarip.uriSimpleRegExp);
        if (matches) {
            var protocol = matches[4] || matches[5] || "";
            var ip = matches[14] || matches[99] || "";
            var domain = matches[12] || matches[172] || matches[184] || matches[191] || matches[677] || "";
            if (domain && !yarip.findTldRegExp.test(domain)) {
                var simple = domain;
                domain = "";
            } else {
                var simple = "";
            }
            var type = domain ? DOMAIN : ip ? IP : simple ? SIMPLE : UNKNOWN;
        } else {
            var protocol = "";
            var ip = "";
            var domain = "";
            var simple = "";
            var type = UNKNOWN;
        }
        var ipArr = [];
        var tldArr = [];
        var domainArr = [];
        switch (type) {
        case IP:
            ipArr = ip.split('.');
            break;
        case SIMPLE:
            domainArr = simple.split('.').reverse();
            break;
        case DOMAIN:
            tldArr = domain.match(yarip.findTldRegExp)[0].split('.');
            domainArr = domain.split('.').reverse();
            break;
        }

        this.obj = {
            type: type,
            protocol: protocol,
            ipArr: ipArr || [],
            tldArr: tldArr || [],
            domainArr: domainArr || []
        }
    }

    this.compare = function(b)
    {
        if (!b) return 1;

        // Comparing types
        var aType = this.obj.type;
        var bType = b.obj.type;
        if (aType < bType) return -1;
        if (aType > bType) return 1;

        // Same types
        switch (aType) {
        case IP:
            // Comparing IPs
            var aIpArr = this.obj.ipArr;
            var bIpArr = b.obj.ipArr;
            if (aIpArr < bIpArr) return -1;
            if (aIpArr > bIpArr) return 1;
            break;

        case SIMPLE:
            // Comparing DOMAINs (SIMPLE)
            var aDomainArr = this.obj.domainArr;
            var bDomainArr = b.obj.domainArr;
            if (aDomainArr < bDomainArr) return -1;
            if (aDomainArr > bDomainArr) return 1;
            break;

        case DOMAIN:
            // Comparing TLDs
            var aTldArr = this.obj.tldArr;
            var bTldArr = b.obj.tldArr;
            if (aTldArr < bTldArr) return -1;
            if (aTldArr > bTldArr) return 1;
            // Comparing DOMAINs
            var aDomainArr = this.obj.domainArr;
            var bDomainArr = b.obj.domainArr;
            if (aDomainArr < bDomainArr) return -1;
            if (aDomainArr > bDomainArr) return 1;
            // Comparing PROTOCOLs
            var aProtocol = this.obj.protocol;
            var bProtocol = b.obj.protocol;
            if (aProtocol < bProtocol) return -1;
            if (aProtocol > bProtocol) return 1;
            break;
        }

        var aName = this.getName();
        var bName = b.getName();
        if (aName < bName) return -1;
        if (aName > bName) return 1;
        return 0;
    }

    this.generateXml = function()
    {
        if (this.temporary) return "";

        var tmpElement = this.elementWhitelist.generateXml() +
            this.elementBlacklist.generateXml() +
            this.elementAttributeList.generateXml() +
            this.elementScriptList.generateXml();
        if (tmpElement != "") {
            tmpElement = "\t\t<element>\n" + tmpElement + "\t\t</element>\n";
        }

        var tmpContentHeader = this.contentRequestHeaderList.generateXml() +
            this.contentResponseHeaderList.generateXml();
        var tmpContent = this.contentWhitelist.generateXml() +
            this.contentBlacklist.generateXml() +
            (tmpContentHeader != "" ? "\t\t\t<header>\n" + tmpContentHeader + "\t\t\t</header>\n" : "") +
            this.contentRedirectList.generateXml();
        if (tmpContent != "") {
            tmpContent = "\t\t<content>\n" + tmpContent + "\t\t</content>\n";
        }

        var tmpPageHeader = this.pageRequestHeaderList.generateXml() +
            this.pageResponseHeaderList.generateXml();
        var tmpPage = this.pageStyleList.generateXml() +
            this.pageScriptList.generateXml() +
            (tmpPageHeader != "" ? "\t\t\t<header>\n" + tmpPageHeader + "\t\t\t</header>\n" : "") +
            this.pageRedirectList.generateXml() +
            this.pageExtensionList.generateXml();
        if (tmpPage != "") {
            tmpPage = "\t\t<page>\n" + tmpPage + "\t\t</page>\n";
        }

        var tmp = tmpElement + tmpContent + tmpPage;
        if (tmp != "") {
            return "\t<page id=\"" + this.id + "\" name=\"" + this.name + "\">\n" + tmp + "\t</page>\n";
        } else return "";
    }

    this.setId(id);
    this.setName(name, true);
}
